#include "Utils.h"
#include "Likelihood.h"
#include "PriorIndependence.h"
#include "Sampler.h"
#include "SamplerFactory.h"
#include "VariableSet.h"

#include "PosteriorSampleHandlerFileDumper.h"

template<unsigned int D>
class GaussianShellsLikelihood : public bcm::Likelihood
{
public:
	Real log_circ(const Real* theta, Real c, Real r, Real w)
	{
		Real f = bcm::rsqrt(2.0 * M_PI * w * w);
		Real d = 0.0;
		for (unsigned int i = 0; i < D; i++) {
			Real thisd;
			if (i == 0) {
				thisd = theta[i] - c;
			} else {
				thisd = theta[i];
			}
			d += thisd * thisd;
		}
		d = sqrt(d);
		Real e = ((d - r) * (d - r)) / (2 * w * w);
		return log(f) - e;
	}

	bool EvaluateLogProbability(size_t threadix, const Real* values, Real& logp)
	{
		static const Real w = 0.1;
		static const Real r = 2.0;
		static const Real c[2] = {  3.5, -3.5 };

		Real logcirc1 = log_circ(values, c[0], r, w);
		Real logcirc2 = log_circ(values, c[1], r, w);
		logp = bcm::logsum(logcirc1, logcirc2);

		return true;
	}
};

int main(int argc, char* argv[])
{
	bcm::logger->SetLogToConsole(bcm::Logger::Info);
	bcm::logger->SetLogToFile(bcm::Logger::Info, "log.txt");
	bcm::logger->LogTime();

	static const size_t ND = 10;

	GaussianShellsLikelihood<ND> likelihood;

	bcm::VariableSet varset;
	for (size_t i = 0; i < ND; i++) {
		std::string name = std::string("theta") + std::to_string((uint64)i);
		varset.AddVariable(name);
	}

	bcm::PriorIndependence prior(&varset);
	for (size_t i = 0; i < ND; i++) {
		std::string name = std::string("theta") + std::to_string((uint64)i);
		prior.SetMarginal(name, bcm::UnivariateMarginal::CreateUniform(-6, 6));
	}

	//bcm::Sampler* sampler = bcm::SamplerFactory::Create("smc", 1);
	bcm::Sampler* sampler = bcm::SamplerFactory::Create("pt", 1);
	sampler->SetLatentVariableSet(&varset);
	sampler->SetReentrantPrior(&prior);

	Real evidence = std::numeric_limits<Real>::quiet_NaN();
	Real evidence_sd = std::numeric_limits<Real>::quiet_NaN();

	bcm::PosteriorSampleHandlerFileDumper dumper;
	dumper.SetFile("samples.tsv");
	dumper.SetVariableSet(&varset);
	dumper.Initialize();
	sampler->SetPosteriorSampleHandler(dumper.GetFunctionPointer(), NULL);

	// Calculate evidence for model 1
	sampler->SetReentrantLikelihood(&likelihood);
	if (!sampler->Initialize()) {
		return -1;
	}
	if (!sampler->Run(&evidence, &evidence_sd)) {
		return -2;
	}
	sampler->Uninitialize();
	
	LOG("");
	LOG("%u-dimensional:", ND);
	LOG("  Log evidence from sampler: %g +/- %g", evidence, evidence_sd);
	LOG("");

	delete sampler;

	return 0;
}
