#include "Utils.h"
#include "LikelihoodDynamicLibrary.h"
#include "OutputValueHandlerFileDumper.h"
#include "PriorIndependence.h"
#include "PriorMarginalGMM.h"
#include "PriorBivariateMarginalGMM.h"
#include "ProgressIndicatorConsole.h"
#include "PosteriorSampleHandlerFileDumper.h"
#include "Sampler.h"
#include "SamplerFactory.h"
#include "VariableSet.h"

#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>
#include <fstream>

static const int version_major = 2;
static const int version_minor = 0;
static const int version_fix = 0;

namespace po = boost::program_options;

int run(const po::variables_map& vm)
{
	bcm::Sampler* sampler = NULL;
	bcm::VariableSet* varset = NULL;
	bcm::PriorIndependence* prior = NULL;
	LikelihoodDynamicLibrary* likelihood = NULL;
	bcm::PosteriorSampleHandlerFileDumper* posterior_dumper;
	bcm::OutputValueHandlerFileDumper* output_dumper;

	try {
		std::string model_name = vm["model"].as<std::string>();
		std::string output_folder = vm["output.folder"].as<std::string>();
		std::string output_format = vm["output.posterior_format"].as<std::string>();
		std::transform(output_format.begin(), output_format.end(), output_format.begin(), ::tolower);
		if (output_format != "bin" && output_format != "tsv") {
			LOGERROR("Unrecognized posterior output format \"%s\"", output_format.c_str());
			return -5;
		}
		printf("Model: %s\n", model_name.c_str());
		printf("Output folder: %s\n", output_folder.c_str());

		// Make sure output folder exists
		std::string output_path = model_name + "/" + output_folder;
		bcm::fix_path(output_path);
		if (!boost::filesystem::is_directory(output_path)) {
			std::string create_path = output_path.substr(0, output_path.size()-1);
			if (!boost::filesystem::create_directories(create_path)) {
				LOGERROR("Could not create output directory.");
				return false;
			}
		}

		// Open up a log file
		std::string log_filename = output_path + "log.txt";
		bcm::logger->SetLogToFile(bcm::Logger::Info, log_filename.c_str());
		bcm::logger->LogTime();
		
		// Determine the number of threads we should use
		size_t numthreads = vm["numthreads"].as<size_t>();
		if (numthreads <= 0) {
			numthreads = boost::thread::hardware_concurrency();
			printf("numthreads <= 0 specified, found %u hardware concurrency\n", (unsigned int)numthreads);
			if (numthreads <= 0) {
				numthreads = 1;
			}
		}

		// Create the sampler
		sampler = bcm::SamplerFactory::Create(vm, numthreads);
		if (!sampler) {
			return -2;
		}
		
		// Variable set
		std::string varset_fn = model_name + "/" + model_name + ".xml";
		varset = new bcm::VariableSet;
		if (!varset->LoadFromXML(varset_fn.c_str())) {
			return -3;
		}
		sampler->SetLatentVariableSet(varset);

		// Prior
		//prior = new bcm::PriorMarginalGMM(varset);
		//prior = new bcm::PriorBivariateMarginalGMM(varset);
		prior = new bcm::PriorIndependence(varset);
		if (!prior->LoadFromXML(varset_fn)) {
			return -3;
		}
		sampler->SetReentrantPrior(prior);

		// Likelihood
		likelihood = new LikelihoodDynamicLibrary(numthreads);
#if PLATFORM_WINDOWS
		std::string dynlib_fn = model_name + "/" + model_name + ".dll";
#else
		std::string dynlib_fn = model_name + "/" + model_name + ".so";
#endif
		if (!likelihood->Load(dynlib_fn)) {
			return -4;
		}
		sampler->SetReentrantLikelihood(likelihood);

		// The samples dumper
		posterior_dumper = new bcm::PosteriorSampleHandlerFileDumper;
		posterior_dumper->SetFile(output_path + "samples." + output_format);
		posterior_dumper->SetVariableSet(varset);
		if (!posterior_dumper->Initialize()) {
			return -6;
		}
		sampler->SetPosteriorSampleHandler(posterior_dumper->GetFunctionPointer(), NULL);

		// Output value dumepr
		output_dumper = new bcm::OutputValueHandlerFileDumper;
		output_dumper->SetFile(output_path + "output_values." + output_format);
		size_t n_outvar = 0;
		if (!likelihood->GetOutput(0, NULL, n_outvar)) {
			n_outvar = 0;
		}
		output_dumper->SetNumValues(n_outvar);
		if (!output_dumper->Initialize()) {
			return -7;
		}
		sampler->SetOutputValuesHandler(output_dumper->GetFunctionPointer(), NULL);
		sampler->SetOutputPath(output_path);
	} catch (const std::exception &e) {
		LOGERROR("Program options error: %s", e.what());
		return -1;
	}

	// Everything has been set, initialize the sampler
	if (!sampler->Initialize()) {
		return -10;
	}

	printf("Starting sampler...\n");
	bcm::ProgressIndicatorConsole pic;
	sampler->SetProgressIndicator(&pic);

	// Run the sampler
	Real evidence = 0.0, evidence_sd = 0.0;
	bool result = sampler->Run(&evidence, &evidence_sd);

	printf("Sampling finished.\n");
	printf("Evidence: %g +/- %g\n", evidence, evidence_sd);

	// Cleanup
	sampler->Uninitialize();
	delete sampler;
	delete posterior_dumper;
	delete output_dumper;
	delete likelihood;
	delete prior;
	delete varset;

	if (result) {
		return 0;
	} else {
		return -20;
	}
}

int main(int argc, char* argv[])
{
	printf("mdlinf version %u.%u.%u\n", version_major, version_minor, version_fix);

	bcm::logger->SetLogToConsole(bcm::Logger::Error);
	
	int result = -1;
	try {
		po::options_description generic_options("Generic options", 120);
		generic_options.add_options()
			("help",															"produce help message")
			("numthreads,j", po::value<size_t>()->default_value(0),				"number of threads to use; set to 0 for using all available hardware concurrency")
		;

		po::options_description config_options("Configuration", 120);
		config_options.add_options()
			("model",										po::value<std::string>()->default_value("model"),				"Name of the model, determines the filename: library will be [x].dll/.so, parameters will be [x].xml")
			("continue",																									"Continue from previous run")
			("runtime_seconds",								po::value<size_t>()->default_value(0),							"Number of seconds to run, set to 0 to run until completion. Otherwise the state reached after this time period will be dumped for the next round.")
			("output.folder",								po::value<std::string>()->default_value("output"),				"Folder in which output files will be generated, subfolder of the model directory.")
			("output.posterior_format",						po::value<std::string>()->default_value("bin"),					"Format of the posterior samples, can be either \"bin\" or \"tsv\".")
		;
		bcm::SamplerFactory::AddOptionsDescription(config_options);

		po::options_description cmdline_only_options("Command-line only options", 120);
		cmdline_only_options.add_options()
			("config_file,c", po::value<std::string>()->default_value("config.txt"),		"Configuration file")
		;
		
		// Add all options together
		po::options_description cmdline_options("Allowed options", 120);
		cmdline_options.add(generic_options).add(cmdline_only_options).add(config_options);

		po::positional_options_description pod;
		pod.add("model", 1);
		
		// Parse command line
		po::variables_map vm;
		po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(pod).run(), vm);
		
		// Handle help command
		if (vm.count("help")) {
			std::cout << cmdline_options << "\n";
			result = 0;
		} else {
			// Parse config file, which may have been specified on the command line, otherwise load the default.
			std::string config_file = vm["config_file"].as<std::string>();
			po::store(po::parse_config_file<char>(config_file.c_str(), config_options), vm);
		
			// Basic parsing finished, give the rest to the main program.
			result = run(vm);
		}
	} catch (po::error &e) {
		LOGERROR("Error processing program options: %s", e.what());
	}

	return result;
}
