#include "Utils.h"
#include "Model.h"
#include "ParseUtilFunctions.h"

#include <fstream>
#include <boost/regex.hpp>

namespace ModelParser {

Model::Model()
	: Evaluation_CurrentIndentation(0)
{
	likelihood = new Likelihood(this);
}

Model::~Model()
{
	delete likelihood;
}

bool Model::Parse(const std::string& inputfn)
{
	boost::smatch modelname_match;
	if (boost::regex_search(inputfn, modelname_match, boost::regex("(\\w+)\\.txt$"))) {
		ModelName = modelname_match[1].str();
	} else {
		LOGERROR("Could not deduce model name from file \"%s\"", inputfn.c_str());
		return false;
	}

	std::ifstream ifs;
	ifs.open(inputfn, std::ifstream::in);
	if (!ifs.is_open()) {
		LOGERROR("Unable to open input file \"%s\"", inputfn.c_str());
		return false;
	}

	ESection current_section = S_Invalid;

	Model model;

	for (unsigned int linenr = 1; ifs.good(); linenr++) {
		std::string line;
		std::getline(ifs, line);

		// Remove comments
		RemoveComments(line);
		boost::trim(line);
		if (line.empty()) {
			continue;
		}

		// New section?
		boost::smatch category_match;
		if (boost::regex_match(line, category_match, boost::regex("^[\\s\\t]{0,}\\[(\\w+)\\][\\s\\t]{0,}$"))) {
			current_section = GetSection(category_match[1].str());
			continue;
		}
		
		// Parse line
		switch(current_section) {
		case S_Constants:
			{
				boost::regex constant_re("(.+?)=(.+?)");
				boost::smatch results;
				if (boost::regex_match(line, results, constant_re)) {
					std::string constant = results[1].str();
					std::string value = results[2].str();
					boost::trim(constant);
					boost::trim(value);

					if (strncmp(constant.c_str(), "real", 4) == 0) {
						constant = std::string("double") + constant.substr(4);
					}

					Constants[constant] = value;
				} else {
					LOGERROR("Error in constant specification at line %u: \"%s\"", linenr, line.c_str());
					return false;
				}
			}
			break;

		case S_Data:
			{
				Data d;
				if (!d.Parse(line, linenr)) {
					return false;
				}
				Datasets.push_back(d);
			}
			break;

		case S_Variables:
			{
				Variable v;
				if (!v.Parse(line, linenr)) {
					return false;
				}
				Variables.push_back(v);
			}
			break;

		case S_Priors:
			{
				boost::regex prior_re("(.+?)=(.+?)");
				boost::smatch results;
				if (boost::regex_match(line, results, prior_re)) {
					std::string name = results[1].str();
					std::string spec = results[2].str();
					boost::trim(name);

					std::string indexing;
					std::string varname;
					if (boost::regex_match(name, results, boost::regex("(\\w+)\\[(.+?)\\]"))) {
						// Indexing
						varname = results[1].str();
						indexing = results[2].str();
					} else {
						// No indexing - leave name as is.
						varname = name;
					}

					Variable* v = FindVariable(varname);
					if (v) {
						if (!v->ParsePrior(line, linenr, spec, indexing)) {
							return false;
						}
					} else {
						LOGERROR("Prior specified for unknown variable \"%s\" at line %u: \"%s\"", name.c_str(), linenr, line.c_str());
						return false;
					}
				}
			}
			break;

		case S_Likelihood:
			if (!likelihood->Parse(line, linenr)) {
				return false;
			}
			break;

		case S_Invalid:
		default:
			LOGERROR("Invalid section");
			return false;
		}
	}

	ifs.close();

	size_t latent_ix = 0;
	for (std::vector<Variable>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		if (vi->IsLatent()) {
			vi->SetLatentIndex(latent_ix);
			latent_ix += vi->GetNumReplicates();
		}
	}

	return true;
}

bool Model::WriteHeader(const std::string& outputpath)
{
	std::string outputfn = outputpath + ModelName + ".h";

	std::ofstream ofs;
	ofs.open(outputfn, std::ofstream::out);
	if (!ofs.is_open()) {
		LOGERROR("Unable to open output file \"%s\"", outputfn.c_str());
		return false;
	}

	std::string class_name = GetClassName();

	ofs << "#pragma once\n";
	ofs << "\n";
	ofs << "class " << class_name << "\n";
	ofs << "{\n";
	ofs << "public:\n";
	ofs << "	bool EvaluateLikelihoodImpl(const double* values, double* llh);\n";
	ofs << "	bool GetOutputImpl(double* output_values, size_t& numvalues);\n";
	ofs << "	\n";
	ofs << "private:\n";
	ofs << "	// Functions\n";

	likelihood->WriteFunctionsHeader(ofs);

	ofs << "	\n";
	ofs << "	// Variables\n";

	for (std::vector<Variable>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		std::string type;
		if		(vi->GetType() == Variable::T_Boolean) type = "bool";
		else if (vi->GetType() == Variable::T_Real) type = "double";
		else {
			LOGERROR("Invalid variable type");
			return false;
		}

		if (vi->GetNumReplicates() == 1) {
			ofs << "	" << type << " " << vi->GetName() << ";\n";
		} else {
			ofs << "	" << type << " " << vi->GetName() << "[" << vi->GetNumReplicates() << "];\n";
		}
	}

	ofs << "\n";
	ofs << "	// Data\n";
	
	for (std::vector<Data>::const_iterator di = Datasets.begin(); di != Datasets.end(); ++di) {
		di->WriteHeader(ofs);
	}

	ofs << "};\n";

	ofs.close();
	return true;
}

bool Model::WriteSource(const std::string& outputpath)
{
	std::string outputfn = outputpath + ModelName + ".cpp";

	std::ofstream ofs;
	ofs.open(outputfn, std::ofstream::out);
	if (!ofs.is_open()) {
		LOGERROR("Unable to open output file \"%s\"", outputfn.c_str());
		return false;
	}

	std::string class_name = GetClassName();
	
	ofs << "#include \"Utils.h\"\n";
	ofs << "#include \"ProbabilityDistributions.h\"\n";
	ofs << "#include \"" << ModelName << ".h\"\n";
	ofs << "\n";
	ofs << "using namespace bcm;\n";
	ofs << "\n";

	ofs << "static const double infinity = std::numeric_limits<double>::infinity();\n";
	for (std::map<std::string, std::string>::const_iterator ci = Constants.begin(); ci != Constants.end(); ++ci) {
		ofs << "static const " << ci->first << " = " << ci->second << ";\n";
	}
	ofs << "\n";

	ofs << class_name << "* models;\n";
	ofs << "\n";

#if PLATFORM_WINDOWS
	ofs << "#include <windows.h>\n";
#endif
	ofs << "#define EXPORT_PREFIX  extern \"C\"\n";

	ofs << "\n";
	ofs << "EXPORT_PREFIX bool AllocateModels(size_t num_models)\n";
	ofs << "{\n";
	ofs << "	models = new " << class_name << "[num_models];\n";
	ofs << "	return (models != NULL);\n";
	ofs << "}\n";
	ofs << "\n";
    ofs << "EXPORT_PREFIX void DeallocateModels()\n";
    ofs << "{\n";
	ofs << "	delete [] models;\n";
    ofs << "}\n";
	ofs << "\n";
	ofs << "EXPORT_PREFIX bool EvaluateLikelihood(size_t model_i, const double* values, double* llh)\n";
	ofs << "{\n";
	ofs << "	return models[model_i].EvaluateLikelihoodImpl(values, llh);\n";
	ofs << "}\n";
	ofs << "\n";
	ofs << "EXPORT_PREFIX bool GetOutput(size_t model_i, double* output_values, size_t& numvalues)\n";
	ofs << "{\n";
	ofs << "	return models[model_i].GetOutputImpl(output_values, numvalues);\n";
	ofs << "}\n";
	ofs << "\n";
	for (std::vector<Data>::const_iterator di = Datasets.begin(); di != Datasets.end(); ++di) {
		di->WriteSource(ofs, GetClassName());
	}

	ofs << "\n";
	likelihood->WriteFunctionsSource(ofs);

	ofs << "\n";
	ofs << "bool " << class_name << "::EvaluateLikelihoodImpl(const double* values, double* llh)\n";
	ofs << "{\n";
	ofs << "	// Set latent variables to the provided values\n";
	for (std::vector<Variable>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		if (vi->IsLatent()) {
			if (vi->GetNumReplicates() == 1) {
				ofs << "\t" << vi->GetName() << " = values[" << vi->GetLatentIndex() << "];\n";
			} else {
				ofs << "\tfor (size_t i = 0; i < " << vi->GetNumReplicates() << "; i++) {\n";
				ofs << "\t\t" << vi->GetName() << "[i] = values[" << vi->GetLatentIndex() << "+i];\n";
				ofs << "\t}\n";
			}
		}
	}
	ofs << "\n";
	ofs << "	// Calculate likelihood\n";
	ofs << "	double logp = 0.0;\n";
	ofs << "\n";

	likelihood->WriteSource(ofs);
	
	size_t numvar = 0;
	for (std::vector<Variable>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		numvar += vi->GetNumReplicates();
	}

	ofs << "\n";
	ofs << "	*llh = logp;\n";
	ofs << "	return true;\n";
	ofs << "}\n";
	ofs << "\n";
	ofs << "bool " << class_name << "::GetOutputImpl(double* output_values, size_t& numvalues)\n";
	ofs << "{\n";
	ofs << "	if (!output_values) {\n";
	ofs << "		numvalues = " << std::to_string((uint64)numvar) << ";\n";
	ofs << "		return true;\n";
	ofs << "	} else {\n";
	ofs << "		if (numvalues < " << std::to_string((uint64)numvar) << ") {\n";
	ofs << "			return false;\n";
	ofs << "		} else {\n";
	ofs << "			size_t i = 0;\n";
	for (std::vector<Variable>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		if (vi->GetNumReplicates() == 1) {
			ofs << "			output_values[i++] = " << vi->GetName() << ";\n";
		} else {
			ofs << "			for (size_t j = 0; j < " << vi->GetNumReplicates() << "; j++) {\n";
			ofs << "				output_values[i++] = " << vi->GetName() << "[j];\n";
			ofs << "			}\n";
		}
	}
	ofs << "			return true;\n";
	ofs << "		}\n";
	ofs << "	}\n";
	ofs << "}\n";
	
	ofs.close();
	
	outputfn = outputpath + ModelName + ".def";
	
	ofs.open(outputfn, std::ofstream::out);
	if (!ofs.is_open()) {
		LOGERROR("Unable to open output file \"%s\"", outputfn.c_str());
		return false;
	}

	ofs << "EXPORTS\n";
	ofs << "	AllocateModels\n";
	ofs << "	DeallocateModels\n";
	ofs << "	EvaluateLikelihood\n";
	ofs << "	GetOutput\n";

	ofs.close();
	return true;
}

bool Model::WriteVariableFile(const std::string& outputpath)
{
	std::string outputfn = outputpath;
	bcm::fix_path(outputfn);
	outputfn += ModelName + ".xml";
	
	size_t noutvar = 0;
	for (std::vector<Variable>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		noutvar += vi->GetNumReplicates();
	}

	std::ofstream ofs;
	ofs.open(outputfn, std::ofstream::out);
	if (!ofs.is_open()) {
		LOGERROR("Unable to open output file \"%s\"", outputfn.c_str());
		return false;
	}
	
	ofs << "<variableset>\n";
	for (std::vector<Variable>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		if (vi->IsLatent()) {
			for (size_t i = 0; i < vi->GetNumReplicates(); i++) {
				ofs << vi->GetVariableFileString(i) + "\n";
			}
		}
	}
	ofs << "	<output_variables>" << std::to_string((uint64)noutvar) << "</output_variables>\n";
	ofs << "</variableset>\n";

	ofs.close();
	return true;
}

bool Model::WriteCMakeFile(const std::string& outputpath, const std::string& bcm_path)
{
	std::string outputfn = outputpath + "CMakeLists.txt";
	
	std::ofstream ofs;
	ofs.open(outputfn, std::ofstream::out);
	if (!ofs.is_open()) {
		LOGERROR("Unable to open output file \"%s\"", outputfn.c_str());
		return false;
	}
	
	ofs << "project(" << ModelName << ")\n";
	ofs << "cmake_minimum_required(VERSION 2.6)\n";
	ofs << "\n";
	ofs << "if(CMAKE_HOST_UNIX)\n";
	ofs << "	set(CMAKE_BUILD_TYPE Release)\n";
	ofs << "	set(CMAKE_CXX_FLAGS \"-O3 -std=c++11 -march=native\")\n";
	ofs << "endif(CMAKE_HOST_UNIX)\n";
	ofs << "\n";
	ofs << "set(Boost_NO_BOOST_CMAKE       TRUE)\n";
	ofs << "set(Boost_USE_STATIC_LIBS        ON)\n";
	ofs << "set(Boost_USE_MULTITHREADED      ON)\n";
	ofs << "set(Boost_USE_STATIC_RUNTIME    OFF)\n";
	ofs << "find_package(Boost COMPONENTS chrono date_time filesystem system thread)\n";
	ofs << "include_directories(" << bcm_path << "src/utils ${Boost_INCLUDE_DIRS})\n";
	ofs << "include_directories(" << bcm_path << "dependencies/eigen-3.2.7)\n";
	ofs << "link_directories(" << bcm_path << "lib)\n";
	ofs << "add_library(" << ModelName << " MODULE " << ModelName << ".cpp " << ModelName << ".h " << ModelName << ".def)\n";
	ofs << "if(CMAKE_HOST_WIN32 OR CMAKE_HOST_APPLE)\n";
	ofs << "	target_link_libraries(" << ModelName << " " << "${Boost_LIBRARIES} debug utilsd optimized utils)\n";
	ofs << "endif(CMAKE_HOST_WIN32 OR CMAKE_HOST_APPLE)\n";
	ofs.close();
	return true;
}

const Variable* Model::FindVariable(const std::string& name) const
{
	for (std::vector<Variable>::const_iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		if (vi->GetName() == name) {
			return &*vi;
		}
	}
	return NULL;
}

const Data* Model::FindData(const std::string& name) const
{
	for (std::vector<Data>::const_iterator di = Datasets.begin(); di != Datasets.end(); ++di) {
		if (di->GetName() == name) {
			return &*di;
		}
	}
	return NULL;
}

std::string Model::GetClassName() const
{
	return std::string("Model_") + ModelName;
}

Variable* Model::FindVariable(const std::string& name)
{
	for (std::vector<Variable>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		if (vi->GetName() == name) {
			return &*vi;
		}
	}
	return NULL;
}

Data* Model::FindData(const std::string& name)
{
	for (std::vector<Data>::iterator di = Datasets.begin(); di != Datasets.end(); ++di) {
		if (di->GetName() == name) {
			return &*di;
		}
	}
	return NULL;
}

Model::ESection Model::GetSection(const std::string& str)
{
	if (strcasecmp(str.c_str(), "constants"	) == 0) return S_Constants;
	if (strcasecmp(str.c_str(), "data"		) == 0) return S_Data;
	if (strcasecmp(str.c_str(), "variables"	) == 0) return S_Variables;
	if (strcasecmp(str.c_str(), "priors"	) == 0) return S_Priors;
	if (strcasecmp(str.c_str(), "likelihood") == 0) return S_Likelihood;
	else return S_Invalid;
}

}
