#include "Utils.h"
#include "ParseUtilFunctions.h"
#include "Variable.h"

#include <boost/regex.hpp>

namespace ModelParser {

Variable::Variable()
	: Type(T_Invalid)
	, Replicates(0)
	, Latent(false)
{
}

Variable::~Variable()
{
}

bool Variable::Parse(const std::string& line, unsigned int linenr)
{
	size_t equals_pos = line.find('=');
	if (equals_pos == std::string::npos) {
		LOGERROR("Syntax error in line %u: \"%s\"", linenr, line.c_str());
		return false;
	}

	Name = line.substr(0, equals_pos);
	boost::trim(Name);
	if (Name.empty()) {
		LOGERROR("Empty variable name in line %u: \"%s\"", linenr, line.c_str());
		return false;
	}

	std::string spec = line.substr(equals_pos+1);
	if (spec.empty()) {
		LOGERROR("Syntax error in line %u: \"%s\"", linenr, line.c_str());
		return false;
	}

	size_t naming_pos = spec.find("=>");
	std::string type;
	std::string naming;
	if (naming_pos == std::string::npos) {
		type = line;
	} else {
		type = spec.substr(0, naming_pos);
		naming = spec.substr(naming_pos+2);
	}

	if (strncmp(type.c_str(), "real", 4)) {
		Type = Variable::T_Real;
	} else if (strncmp(type.c_str(), "boolean", 7)) {
		Type = Variable::T_Boolean;
	} else {
		LOGERROR("Unknown variable type in line %u: \"%s\"", linenr, line.c_str());
		return false;
	}

	std::string array_str;
	if (GetArrayStr(type, array_str)) {
		if (array_str.empty()) {
			Replicates = 1;
		} else { 
			Replicates = (size_t)atoi(array_str.c_str());
		}
	} else {
		LOGERROR("Syntax error in line %u: \"%s\"", linenr, line.c_str());
		return false;
	}

	if (!naming.empty()) {
		std::string names = naming.substr(naming.find("{")+1, naming.find("}") - (naming.find("{")+1));
		bcm::tokenize(names, ReplicateNames, ",");
		if (ReplicateNames.size() != Replicates) {
			LOGERROR("Not enough names for replicates on line %u: \"%s\"", linenr, line.c_str());
			return false;
		}
		for (std::vector<std::string>::iterator rni = ReplicateNames.begin(); rni != ReplicateNames.end(); ++rni) {
			boost::trim(*rni);
		}
		for (std::vector<std::string>::iterator rni = ReplicateNames.begin(); rni != ReplicateNames.end(); ++rni) {
			boost::trim(*rni);
			RemoveEnclosingCharacters(*rni, '"', '"');
		}
	} else {
		ReplicateNames.resize(Replicates);
		for (size_t i = 0; i < Replicates; i++) {
			ReplicateNames[i] = std::to_string((uint64)i);
		}
	}

	return true;
}

bool Variable::ParsePrior(const std::string& line, unsigned int linenr, const std::string& prior, std::string index)
{
	Latent = true;
	Priors.resize(Replicates);

	Prior p;

	boost::regex re("(\\w+)\\((.+?)\\)");
	boost::smatch results;
	if (boost::regex_search(prior, results, re)) {
		p.PriorType = results[1].str();
		std::string contents = results[2].str();
		std::vector<std::string> params;
		bcm::tokenize(contents, params, ",");
		for (std::vector<std::string>::iterator pi = params.begin(); pi != params.end(); ++pi) {
			std::string str = *pi;
			boost::trim(str);
			
			boost::regex param_re("(\\w+)[\\s\\t]{0,}=[\\s\\t]{0,}([\\w\\.-]+)");
			boost::smatch param_results;
			if (boost::regex_match(str, param_results, param_re)) {
				p.PriorParams[param_results[1].str()] = param_results[2].str();
			} else {
				LOGERROR("Syntax error for prior of parameter \"%s\" in string \"%s\" at line %u: \"%s\"", Name.c_str(), str.c_str(), linenr, line.c_str());
				return false;
			}
		}
	} else {
		LOGERROR("Syntax error for prior of parameter \"%s\" at line %u: \"%s\"", Name.c_str(), linenr, line.c_str());
		return false;
	}

	if (index == "") {
		for (size_t i = 0; i < Replicates; i++) {
			Priors[i] = p;
		}
	} else {
		size_t ix = (size_t)atoi(index.c_str());
		if (index == std::to_string((uint64)ix)) {
			if (ix < Replicates) {
				Priors[ix] = p;
			} else {
				LOGERROR("Error for prior of parameter \"%s\", index \"%s\" looks like a number, but there are only %u replicates, at line %u: \"%s\"", Name.c_str(), index.c_str(), Replicates, linenr, line.c_str());
				return false;
			}
		} else {
			boost::smatch results;
			if (boost::regex_match(index, results, boost::regex("\\\"(.+?)\\\""))) {
				std::string index_name = results[1].str();
				bool found = false;
				for (size_t i = 0; i < Replicates; i++) {
					if (ReplicateNames[i] == index_name) {
						Priors[i] = p;
						found = true;
						break;
					}
				}

				if (!found) {
					LOGERROR("Error for prior of parameter \"%s\", index \"%s\" not found, at line %u: \"%s\"", Name.c_str(), index_name.c_str(), linenr, line.c_str());
					return false;
				}
			} else {
				LOGERROR("Error for prior of parameter \"%s\", don't know what to do with index \"%s\", at line %u: \"%s\"", Name.c_str(), index.c_str(), linenr, line.c_str());
				return false;
			}
		}
	}

	return true;
}

bool Variable::GetReplicateIndex(const std::string& replicate_name, size_t& ix) const
{
	ix = 0xffffffff;
	for (size_t i = 0; i < ReplicateNames.size(); i++) {
		if (replicate_name == ReplicateNames[i]) {
			ix = i;
			return true;
		}
	}
	return false;
}

std::string Variable::GetVariableFileString(size_t ix) const
{
	std::string str("\t<variable name=\"");
	str += Name;
	if (Replicates > 1) {
		str += "[";
		str += ReplicateNames[ix];
		str += "]";
	}
	str += "\" distribution=\"";
	str += Priors[ix].PriorType;
	str += "\"";
	for (std::map<std::string, std::string>::const_iterator ppi = Priors[ix].PriorParams.begin(); ppi != Priors[ix].PriorParams.end(); ++ppi) {
		str += " ";
		str += ppi->first;
		str += "=\"";
		str += ppi->second;
		str += "\"";
	}
	str += "/>";
	return str;
}

void Variable::SetLatentIndex(size_t start)
{
	ASSERT(IsLatent());
	LatentIndex = start;
}

}
