#include "Utils.h"
#include "Prior.h"
#include "NestedSamplingStrategyMultiNest.h"
#include "NestedSamplingUtils.h"
#include "VariableSet.h"

#include <Eigen/Eigenvalues>

namespace bcm {

NestedSamplingStrategyMultiNest::Cluster::Cluster(size_t nvar)
	: nvar(nvar)
	, mean(VectorReal::Zero(nvar))
	, covariance(MatrixReal::Zero(nvar, nvar))
	, inv_covariance(MatrixReal::Zero(nvar, nvar))
	, transformation(MatrixReal::Zero(nvar, nvar))
	, scaling(1.0)
{
	parent = NULL;
	children[0] = NULL;
	children[1] = NULL;
}

NestedSamplingStrategyMultiNest::Cluster::~Cluster()
{
	safe_delete(children[0]);
	safe_delete(children[1]);
}

void NestedSamplingStrategyMultiNest::Cluster::CalculateMean(const std::vector<VectorReal>& unit_samples)
{
	mean.setZero();
	for (size_t k = 0; k < sample_ixs.size(); k++) {
		const VectorReal& s = unit_samples[sample_ixs[k]];
		const Real one_over_kp1 = 1.0 / (k + 1.0);
		for (size_t i = 0; i < nvar; i++) {
			mean(i) += (s(i) - mean(i)) * one_over_kp1;
		}
	}
}

void NestedSamplingStrategyMultiNest::Cluster::CalculateCovariance(const std::vector<VectorReal>& unit_samples)
{
	covariance.setZero();
	for (size_t k = 0; k < sample_ixs.size(); k++) {
		const VectorReal& s = unit_samples[sample_ixs[k]];
		Real one_over_kp1 = 1.0 / (k + 1.0);
		for (size_t i = 0; i < nvar; i++) {
			for (size_t j = 0; j < nvar; j++) {
				Real d1 = s(i) - mean(i);
				Real d2 = s(j) - mean(j);
				covariance(i,j) += (d1 * d2 - covariance(i,j)) * one_over_kp1;
			}
		}
	}
	covariance *= sample_ixs.size() / ((Real)sample_ixs.size() - 1.0);

	if (covariance(0,0) != covariance(0,0)) {
		LOG("Error");
	}
}

void NestedSamplingStrategyMultiNest::Cluster::CalculateBoundingEllipsoid(const std::vector<VectorReal>& unit_samples)
{
	CalculateMean(unit_samples);
	CalculateCovariance(unit_samples);

	// Calculate ellipsoid orientation
	MatrixReal eigvec, eigval;
	bcm::NestedSamplingUtils::EigenDecomposition(covariance, eigval, eigvec, inv_covariance);
	
	for (size_t j = 1; j < nvar; j++) {
		if (eigval(j,j) <= 0) {
			eigval(j,j) = 1e-9;
		}
	}

	// Calculate ellipsoid scaling so that it encloses all points
	VectorReal centered(nvar);
	VectorReal optimize(nvar);
	Real fk = 0.0;
	for (size_t i = 0; i < sample_ixs.size(); i++) {
		const VectorReal& s = unit_samples[sample_ixs[i]];
		centered = s - mean;
		
		for (size_t j = 0; j < nvar; j++) {
			optimize(j) = 0;
			for (size_t k = 0; k < nvar; k++) {
				optimize(j) += inv_covariance(j,k) * centered(k);
			}
		}
		Real d = optimize.dot(centered);
		if (d < 9.0) {
			fk = std::max(fk, d);
		} else {
			// More than 3 standard deviations away - ignore as outlier
			// TODO - should probably remove outliers before calculating the covariance.
			// The fortran version of MultiNest seems to do something like that as well, but the Matlab
			// version doesn't have outlier removal.
		}
	}

	MatrixReal scaled = fk * covariance;
	Eigen::EigenSolver<MatrixReal> es;
	es.compute(scaled);
	eigvec = es.eigenvectors().real();
	eigval = MatrixReal::Zero(scaled.rows(), scaled.rows());
	eigval.diagonal() = es.eigenvalues().real();
	
	// Calculate volume of ellipsoid
	if (nvar <= 20) {
		Real determinant = 1.0;
		for (size_t j = 0; j < nvar; j++) {
			determinant *= eigval(j,j);
		}
		volume = sqrt(determinant);
		if (nvar % 2 == 0) {
			for (size_t i = 2; i <= nvar; i += 2) {
				volume *= 2.0 * M_PI / (Real)i;
			}
		} else {
			volume *= 2.0;
			for (size_t i = 3; i <= nvar; i += 2) {
				volume *= 2.0 * M_PI / (Real)i;
			}
		}
	} else {
		// Log scale to prevent underflow
		Real determinant = 0.0;
		for (size_t j = 0; j < nvar; j++) {
			determinant += log(eigval(j,j));
		}
		volume = 0.5 * determinant;
		if (nvar % 2 == 0) {
			for (size_t i = 2; i <= nvar; i += 2) {
				volume += log(2.0 * M_PI / (Real)i);
			}
		} else {
			volume *= 2.0;
			for (size_t i = 3; i <= nvar; i += 2) {
				volume += log(2.0 * M_PI / (Real)i);
			}
		}
		volume = exp(volume);
	}
	ASSERT(volume == volume);

	// Finally the transformation matrix
	for (size_t i = 0; i < nvar; i++) {
		transformation.col(i) = eigvec.col(i) * sqrt(eigval(i,i));
	}
}

void NestedSamplingStrategyMultiNest::Cluster::Recalculate(const std::vector<VectorReal>& unit_samples, Real minimum_volume, RNG* rng)
{
	static const size_t MAX_K_MEANS_STEPS = 3;
	static const size_t MAX_IMPROVEMENT_STEPS = 20;

	// We need at least two times D+1 points to do meaningful clustering.
	// But set it to 2*D to get a bit more robust ellipsoids
	const size_t min_n = nvar + 1;
	const size_t n = sample_ixs.size();
	if (n < min_n * 2) {
		return;
	}

	children[0] = new Cluster(nvar);
	children[1] = new Cluster(nvar);
	children[0]->sample_ixs.reserve(n);
	children[1]->sample_ixs.reserve(n);

	Real VS1 = 0.0, VS2 = 0.0;

	bool success = false;
	for (size_t overall_tries = 0; overall_tries < MAX_K_MEANS_STEPS; overall_tries++) {
		// 3. Perform K-means clustering
		VectorReal centroid[2];
		std::vector<char> assignment(n, -1);
		size_t assign_count[2] = { 0, 0 };
		bool have_two_clusters = true;

		// Initial points using K-means++.
		size_t pt1 = rng->GetUniformInt((unsigned long)n);
		centroid[0] = unit_samples[sample_ixs[pt1]];

		size_t pt2 = n-1;
		std::vector<Real> distance(n, 0.0);
		Real total = 0.0;
		for (size_t i = 0; i < n; i++) {
			const VectorReal& x = unit_samples[sample_ixs[i]];
			distance[i] = bcm::NestedSamplingUtils::EuclideanDistanceSq(centroid[0], x);
			total += distance[i];
		}
		Real p = rng->GetReal();
		Real t = 0;
		total = 1.0 / total;
		for (size_t i = 0; i < n - 1; i++) {
			t += distance[i] * total;
			if (p < t) {
				pt2 = i;
				break;
			}
		}
		centroid[1] = unit_samples[sample_ixs[pt2]];

		for (size_t step = 0; step < MAX_K_MEANS_STEPS; step++) {
			bool moved = false;

			// Assign to nearest centroid
			for (size_t i = 0; i < n; i++) {
				const size_t ix = sample_ixs[i];
				const VectorReal& s = unit_samples[ix];

				Real min_dist = -std::numeric_limits<Real>::infinity();
				Real d1 = bcm::NestedSamplingUtils::EuclideanDistanceSq(s, centroid[0]);
				Real d2 = bcm::NestedSamplingUtils::EuclideanDistanceSq(s, centroid[1]);
				if (d1 < d2) {
					if (assignment[i] != 0) {
						assignment[i] = 0;
						moved = true;
					}
				} else {
					if (assignment[i] != 1) {
						assignment[i] = 1;
						moved = true;
					}
				}
			}

			if (!moved) {
				//LOG("K-means clustering converged in %u steps", step);
				break;
			}

			// Calculate centroid
			centroid[0].setZero();
			centroid[1].setZero();
			assign_count[0] = 0;
			assign_count[1] = 0;
			for (size_t i = 0; i < n; i++) {
				const VectorReal& s = unit_samples[sample_ixs[i]];
				const char c = assignment[i];
				for (size_t j = 0; j < nvar; j++) {
					centroid[c](j) += s(j);
				}
				assign_count[c]++;
			}
			centroid[0] *= 1.0 / (Real)assign_count[0];
			centroid[1] *= 1.0 / (Real)assign_count[1];
		}
	
		// Check whether the starting clusters are beig enough to calculate a bounding ellipsoid
		if (assign_count[0] < min_n || assign_count[1] < min_n) {
			// No, try the k-means again.
			continue;
		}

		// Do the initial split according to the k-means result.
		for (size_t i = 0; i < n; i++) {
			const char c = assignment[i];
			children[c]->sample_ixs.push_back(sample_ixs[i]);
		}

		// Mahalanobis improvement.
		Real VS = minimum_volume;
		VectorReal tmp1(nvar), tmp2(nvar);
		for (size_t step = 0; step < MAX_IMPROVEMENT_STEPS; step++) {
			if (assign_count[0] < min_n || assign_count[1] < min_n) {
				// Can't calculate volume anymore, bail out (the subdivision will be discarded).
				//LOG("Mahalanobis failed: not enough points for volume calculation, after %u steps", step);
				break;
			}

			// Algorithm step 4. Calculate bounding ellipsoids for both sets
			children[0]->CalculateBoundingEllipsoid(unit_samples);
			children[1]->CalculateBoundingEllipsoid(unit_samples);

			// Algorithm step 5. Enlarge bound so that V(Ek) >= V(Sk)
			size_t n1 = children[0]->sample_ixs.size();
			size_t n2 = children[1]->sample_ixs.size();
			VS1 = VS * n1 / (Real)n;
			VS2 = VS * n2 / (Real)n;
			if (children[0]->volume < VS1) {
				children[0]->scaling *= pow(VS1 / children[0]->volume, 2.0 / nvar);
				children[0]->volume = VS1;
			} else {
				children[0]->scaling = 1.0;
			}
			if (children[1]->volume < VS2) {
				children[1]->scaling *= pow(VS2 / children[1]->volume, 2.0 / nvar);
				children[1]->volume = VS2;
			} else {
				children[1]->scaling = 1.0;
			}

			// Algorithm step 6-8. 
			bool moved = false;
			assign_count[0] = 0;
			assign_count[1] = 0;
			for (size_t i = 0; i < n; i++) {
				const size_t ix = sample_ixs[i];
				const VectorReal& s = unit_samples[ix];

				//Real d1 = bcm::NestedSamplingUtils::MahalanobisDistance(s, children[0]->inv_covariance, children[0]->mean, children[0]->scaling);
				//Real d2 = bcm::NestedSamplingUtils::MahalanobisDistance(s, children[1]->inv_covariance, children[1]->mean, children[1]->scaling);
				Real d1 = bcm::NestedSamplingUtils::MahalanobisDistanceSqFast(s, children[0]->inv_covariance, children[0]->mean, children[0]->scaling, tmp1, tmp2);
				Real d2 = bcm::NestedSamplingUtils::MahalanobisDistanceSqFast(s, children[1]->inv_covariance, children[1]->mean, children[1]->scaling, tmp1, tmp2);

				//Real h1 = children[0]->volume * d1 / VS1;
				//Real h2 = children[1]->volume * d2 / VS2;
				Real h1 = d1;
				Real h2 = d2;
		
				if (h1 < h2) {
					assign_count[0]++;
					if (assignment[i] == 1) {
						assignment[i] = 0;
						moved = true;
					}
				} else {
					assign_count[1]++;
					if (assignment[i] == 0) {
						assignment[i] = 1;
						moved = true;
					}
				}
			}
		
			if (moved) {
				children[0]->sample_ixs.clear();
				children[1]->sample_ixs.clear();
				for (size_t i = 0; i < n; i++) {
					const char c = assignment[i];
					children[c]->sample_ixs.push_back(sample_ixs[i]);
				}
			} else {
				//LOG("Mahalanobis improvement converged in %u steps", step);
				success = true;
				break;
			}
		}

		if (success) {
			bool accept_division = true;

			// Each cluster must have at least min_n samples.
			if (children[0]->sample_ixs.size() < min_n || children[1]->sample_ixs.size() < min_n) {
				accept_division = false;
			} else {
				// The two new clusters together must be smaller than the present clust
				if (children[0]->volume + children[1]->volume >= volume) {
					accept_division = false;
				}

				// However, if the volume is still big, we accept the division anyway.
				if (volume > 2.0 * VS) {
					accept_division = true;
				}
			}
			
			if (accept_division) {
				//LOG("Success after %u tries", overall_tries);
				break;
			} else {
				// Try again.
				success = false;
			}
		}

		children[0]->sample_ixs.clear();
		children[1]->sample_ixs.clear();
	}

	//LOG("Volume=%g, child1=%g, child2=%g, n1=%u, n2=%u, accept=%d", volume, children[0]->volume, children[1]->volume, children[0]->sample_ixs.size(), children[1]->sample_ixs.size(), (int)success);

	if (success) {
		children[0]->Recalculate(unit_samples, VS1, rng);
		children[1]->Recalculate(unit_samples, VS2, rng);
	} else {
		safe_delete(children[0]);
		safe_delete(children[1]);
	}
}

void NestedSamplingStrategyMultiNest::Cluster::Rescale(Real minimum_volume, Real n)
{
	// Here we use samples.size(), because it reflects the removed/added samples in between Recalculate()s.
	size_t nk = samples.size();
	Real VS = minimum_volume * nk / n;
	if (volume < VS) {
		scaling *= pow(VS / volume, 2.0 / (Real)nvar);
		volume = VS;
	}
}

NestedSamplingStrategyMultiNest::NestedSamplingStrategyMultiNest(SamplerNested* sampler)
	: NestedSamplingStrategy(sampler)
	, enlargement(1.1)
	, full_update_steps(population_size/4)
	, root(new Cluster(num_variables))
{
	unit_samples.resize(population_size, VectorReal::Zero(num_variables));
	root->sample_ixs.reserve(population_size);
}

NestedSamplingStrategyMultiNest::~NestedSamplingStrategyMultiNest()
{
	delete root;
}

bool NestedSamplingStrategyMultiNest::Initialize(Real prior_volume_estimate, RNG* rng)
{
	// Convert samples from original space to unit hypercube
	bool result = true;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++si) {
		result &= bcm::NestedSamplingUtils::ConvertToUnitSpace((*si)->variable_values, (*si)->unit_values, prior);
	}
	if (!result) {
		return false;
	}

	Real minimum_volume = prior_volume_estimate * enlargement;

	// Calculate initial bounding ellipsoid
	result = Recalculate(minimum_volume, rng);
	if (!result) {
		return false;
	}

	steps_since_update = 0;
	return true;
}

bool NestedSamplingStrategyMultiNest::UpdateRemoveSample(const Sample* removed_sample, Real prior_volume_estimate, bool& reset, RNG* rng)
{
	// Remove sample from cluster
	bool found = false;
	for (std::vector<Cluster*>::iterator ci = leafs.begin(); ci != leafs.end(); ++ci) {
		Cluster* c = *ci;
		for (std::vector<const Sample*>::iterator si = c->samples.begin(); si != c->samples.end(); ++si) {
			if (*si == removed_sample) {
				found = true;
				c->samples.erase(si);
				break;
			}
		}
		if (found) {
			break;
		}
	}

	// Update the bounds
	Real minimum_volume = prior_volume_estimate * enlargement;
	Real fs = 0.0;
	for (std::vector<Cluster*>::iterator ci = leafs.begin(); ci != leafs.end(); ++ci) {
		Cluster* c = *ci;
		c->Rescale(minimum_volume, static_cast<Real>(population_size - 1));
		fs += c->volume;
	}

	if (steps_since_update >= full_update_steps) {
		LOG("Recalculating after %u steps", steps_since_update);
		steps_since_update = 0;
		if (!Recalculate(minimum_volume, rng)) {
			return false;
		}
		reset = true;
	} else {
		steps_since_update++;
	}

	return true;
}

bool NestedSamplingStrategyMultiNest::UpdateAddSample(const Sample* added_sample, RNG* rng)
{
	if (!bcm::NestedSamplingUtils::ConvertToUnitSpace(added_sample->variable_values, added_sample->unit_values, prior)) {
		return false;
	}

	if (added_sample->sampled_from_cluster != std::numeric_limits<size_t>::max()) {
		Cluster* c = leafs[added_sample->sampled_from_cluster];
		c->samples.push_back(added_sample);
	}

	return true;
}

bool NestedSamplingStrategyMultiNest::GetNewSample(Real min_llh, Sample* sample, RNG* rng, size_t threadix) const
{
	VectorReal newsample(num_variables);
	VectorReal tmp1(num_variables), tmp2(num_variables);

	while (1) {
		// Pick one of the ellipsoids proportional to their volume.
		Real p = rng->GetReal();
		size_t ell_ix = leafs.size() - 1;
		Real t = 0;
		for (size_t i = 0; i < leafs.size() - 1; i++) {
			const Cluster* c = leafs[i];
			t += c->volume / total_volume;
			if (p < t) {
				ell_ix = i;
				break;
			}
		}
		const Cluster* cluster = leafs[ell_ix];
		sample->sampled_from_cluster = ell_ix;

		// Generate a sample from this ellipsoid.
		bcm::NestedSamplingUtils::GeneratePointInUnitSphere(newsample, rng);
		newsample = cluster->scaling * cluster->transformation * newsample + cluster->mean;

		// Make sure it's inside the unit hypercube
#if 1
		bool in_unit_hypercube = true;
		for (size_t i = 0; i < num_variables; i++) {
			if (newsample(i) < 0.0 || newsample(i) > 1.0) {
				in_unit_hypercube = false;
				break;
			}
		}
		if (!in_unit_hypercube) {
			continue;
		}
#else
		// It would be nice if we can reflect across the boundaries, but I don't
		// think the volume calculations are still correct in that case
		for (size_t i = 0; i < num_variables; i++) {
			Real nx = newsample(i);
			while (1) {
				if (nx < 0.0) {
					nx = -nx;
				} else if (nx > 1.0) {
					nx = 1.0 - (nx - 1.0);
				} else {
					break;
				}
			}
			newsample(i) = nx;
		}
#endif

		// Find out in how many ellipsoids this sample lies.
		size_t count = 0;
		for (size_t i = 0; i < leafs.size(); i++) {
			const Cluster* c = leafs[i];
			if (bcm::NestedSamplingUtils::PointInEllipsoid(newsample, c->inv_covariance, c->mean, c->scaling)) {
				count++;
			}
		}

		// Accept with probability inversely proportional to the number of ellipsoids the point is in, to account for overlap.
		if (count > 1) {
			Real p = rng->GetReal();
			if (p > 1.0 / (Real)count) {
				continue;
			} else {
				break;
			}
		} else {
			break;
		}
	}

	// Convert from unit hypercube back to the original space
	bcm::NestedSamplingUtils::ConvertToOriginalSpace(newsample, sample->variable_values, prior);

	if (!EvaluatePriorLikelihood(threadix, sample->variable_values.data(), sample->lprior, sample->llh)) {
		return false;
	}
	if (sample->lprior == -std::numeric_limits<Real>::infinity()) {
		sample->lposterior = sample->lprior;
	} else {
		if (!GetOutputValues(threadix, sample->output_values.data())) {
			return false;
		}
		sample->lposterior = sample->lprior + sample->llh;
	}

	return true;
}

bool NestedSamplingStrategyMultiNest::Recalculate(Real minimum_volume, RNG* rng)
{

	// Clean up previous structure
	safe_delete(root->children[0]);
	safe_delete(root->children[1]);

	// Put all points in the root cluster
	root->samples.clear();
	root->sample_ixs.clear();
	unit_samples.clear();
	size_t i = 0;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++i, ++si) {
		root->samples.push_back(*si);
		root->sample_ixs.push_back(i);
		unit_samples.push_back((*si)->unit_values);
	}
#if 0
	FILE* file = fopen("samples.tsv", "wb");
	for (std::vector<VectorReal>::iterator si = unit_samples.begin(); si != unit_samples.end(); ++si) {
		const VectorReal& s = *si;
		fprintf(file, "%g\t%g\n", s(0), s(1));
	}
	fclose(file);
#endif

	// Algorithm step 1. Calculate root bounding ellipsoid
	root->CalculateBoundingEllipsoid(unit_samples);

	// Algorithm step 2. Enlarge.
	if (root->volume < minimum_volume) {
		root->scaling *= pow(minimum_volume / root->volume, 2.0 / (Real)root->nvar);
		root->volume = minimum_volume;
	}
	
	// Recursive subdivision
	root->Recalculate(unit_samples, minimum_volume, rng);

	// Make a convenience list of all the leafs, and calculate total volume.
	leafs.clear();
	total_volume = 0.0;
	GatherLeafs(root);

	// Repopulate the samples vectors for the clusters
	i = 0;
	for (std::list<Sample*>::const_iterator si = sampling_population.cbegin(); si != sampling_population.cend(); ++i, ++si) {
		for (std::vector<Cluster*>::iterator ci = leafs.begin(); ci != leafs.end(); ++ci) {
			Cluster* c = *ci;
			if (std::find(c->sample_ixs.begin(), c->sample_ixs.end(), i) != c->sample_ixs.end()) {
				c->samples.push_back(*si);
				break;
			}
		}
	}

	LOG("Updated ellipsoidal clustering; nclust=%u, total volume=%g (minimum volume=%g)", leafs.size(), total_volume, minimum_volume);

#if 0
	for (size_t i = 0; i < leafs.size(); i++) {
		const Cluster* c = leafs[i];
		char filename[256];
		sprintf(filename, "samples%u.csv", (unsigned int)i);
	
		FILE* file = fopen(filename, "w");
		for (size_t k = 0; k < c->sample_ixs.size(); k++) {
			const VectorReal& s = unit_samples[c->sample_ixs[k]];
			for (size_t j = 0; j < num_variables; j++) {
				fprintf(file, "%g%s", s(j), (j < num_variables-1) ? "," : "\n");
			}
		}
		fclose(file);
	}
#endif

	return true;
}

void NestedSamplingStrategyMultiNest::GatherLeafs(Cluster* cluster)
{
	if (cluster->children[0] != NULL) {
		ASSERT(cluster->children[1] != NULL);
		GatherLeafs(cluster->children[0]);
		GatherLeafs(cluster->children[1]);
	} else {
		total_volume += cluster->volume;
		leafs.push_back(cluster);
	}
}

}
