#include "Utils.h"
#include "NestedSamplingUtils.h"
#include "Prior.h"

namespace bcm {
namespace NestedSamplingUtils {

bool ConvertToUnitSpace(const VectorReal& original, VectorReal& unit, const Prior* prior)
{
	for (size_t i = 0; i < (size_t)original.size(); i++) {
		Real p;
		bool result = prior->EvaluateMarginalCDF(i, original(i), p);
		if (result) {
			unit(i) = p;
		} else {
			return false;
		}
	}
	return true;
}

bool ConvertToOriginalSpace(const VectorReal& unit, VectorReal& original, const Prior* prior)
{
	for (size_t i = 0; i < (size_t)unit.size(); i++) {
		Real q;
		bool result = prior->EvaluateMarginalQuantile(i, unit(i), q);
		if (result) {
			original(i) = q;
		} else {
			return false;
		}
	}
	return true;
}

void EigenDecomposition(const MatrixReal& covariance, MatrixReal& eigval, MatrixReal& eigvec, MatrixReal& inv_covariance)
{
	Eigen::EigenSolver<MatrixReal> es;
	es.compute(covariance);

	eigvec = es.eigenvectors().real();
	eigval = MatrixReal::Zero(covariance.rows(), covariance.rows());
	eigval.diagonal() = es.eigenvalues().real();
	
	MatrixReal inv_eigval = MatrixReal::Zero(covariance.rows(), covariance.rows());
	for (size_t i = 0; i < (size_t)covariance.rows(); i++) {
		inv_eigval(i,i) = 1.0 / es.eigenvalues().real()(i);
	}

	inv_covariance = eigvec * inv_eigval * eigvec.transpose();
}

void GeneratePointInUnitSphere(VectorReal& sample, RNG* rng)
{
	ASSERT(sample.size() > 0);

	for (size_t i = 0; i < (size_t)sample.size(); i++) {
		sample(i) = rng->GetNormal(0, 1);
	}
	
	Real invnorm = 0.0;
	for (size_t i = 0; i < (size_t)sample.size(); i++) {
		invnorm += sample(i) * sample(i);
	}
	invnorm = rsqrt(invnorm);
	
	Real scale = pow(rng->GetReal(), 1.0 / (Real)sample.size()) * invnorm;
	sample *= scale;
}

Real EuclideanDistanceSq(const VectorReal& p1, const VectorReal& p2)
{
	ASSERT(p1.size() == p2.size());

	Real distanceSq = 0;
	for (size_t i = 0; i < (size_t)p1.size(); i++) {
		Real d = p1(i) - p2(i);
		distanceSq += d * d;
	}
	return distanceSq;
}

Real MahalanobisDistance(const VectorReal& p, const MatrixReal& inv_covariance, const VectorReal& center, const Real scale)
{
	VectorReal centered = p - center;
	VectorReal optimize(p.size());
	for (size_t j = 0; j < (size_t)p.size(); j++) {
		optimize(j) = 0;
		for (size_t k = 0; k < (size_t)p.size(); k++) {
			optimize(j) += scale * inv_covariance(j,k) * centered(k);
		}
	}
	Real k = optimize.dot(centered);
	return sqrt(k);
}

Real MahalanobisDistanceSqFast(const VectorReal& p, const MatrixReal& inv_covariance, const VectorReal& center, const Real scale, VectorReal& tmp1, VectorReal& tmp2)
{
	Real inv_scale = 1.0 / scale;
	//Real scaleSq = scale * scale;
	tmp1.noalias() = p - center;
	for (size_t j = 0; j < (size_t)p.size(); j++) {
		tmp2(j) = 0;
		for (size_t k = 0; k < (size_t)p.size(); k++) {
			tmp2(j) += inv_scale * inv_covariance(j,k) * tmp1(k);
		}
	}
	Real k = tmp2.dot(tmp1);
	return k;
}

bool PointInEllipsoid(const VectorReal& p, const MatrixReal& inv_covariance, const VectorReal& center, const Real scale)
{
	VectorReal centered = p - center;
	VectorReal optimize(p.size());
	for (size_t j = 0; j < (size_t)p.size(); j++) {
		optimize(j) = 0;
		for (size_t k = 0; k < (size_t)p.size(); k++) {
			optimize(j) += inv_covariance(j,k) * centered(k);
		}
	}
	Real k = optimize.dot(centered);

	if (k - scale * scale < 1e-9) {
		return true;
	} else {
		return false;
	}
}

bool PointInEllipsoidFast(const VectorReal& p, const MatrixReal& inv_covariance, const VectorReal& center, const Real scale, VectorReal& tmp1, VectorReal& tmp2)
{
	Real inv_scale = 1.0 / scale;
	tmp1.noalias() = p - center;
	for (size_t j = 0; j < (size_t)p.size(); j++) {
		tmp2(j) = 0;
		for (size_t k = 0; k < (size_t)p.size(); k++) {
			tmp2(j) += inv_scale * inv_covariance(j,k) * tmp1(k);
		}
	}
	Real k = tmp2.dot(tmp1);

	if (k < 1.0 + 1e-9) {
		return true;
	} else {
		return false;
	}
}

}
}
