#include "Utils.h"
#include "PosteriorSampleHandlerFileDumper.h"
#include "VariableSet.h"

namespace bcm {

PosteriorSampleHandlerFileDumper::PosteriorSampleHandlerFileDumper(size_t buffer_size)
	: varset(NULL)
	, Initialized(false)
	, output_buffer_ix(0)
{
	output_buffer.resize(buffer_size);
}

PosteriorSampleHandlerFileDumper::~PosteriorSampleHandlerFileDumper()
{
	if (Initialized) {
		Flush();
	}
}

void PosteriorSampleHandlerFileDumper::SetFile(const std::string& filename)
{
	this->filename = filename;
	
	std::string extension;
	size_t extension_pos = filename.find_last_of(".");
	if (extension_pos != std::string::npos) {
		extension = filename.substr(extension_pos + 1);
		std::transform(extension.begin(), extension.end(), extension.begin(), ::tolower);
		if (extension == "tsv" || extension == "txt") {
			binary = false;
		} else {
			binary = true;
		}
	} else {
		binary = true;
	}
}

void PosteriorSampleHandlerFileDumper::SetVariableSet(const VariableSet* varset)
{
	this->varset = varset;
	for (size_t i = 0; i < output_buffer.size(); i++) {
		output_buffer[i].values.resize(varset->GetNumVariables());
	}
}

bool PosteriorSampleHandlerFileDumper::Initialize()
{
	if (filename.empty()) {
		LOGERROR("Output filename has not been specified");
		return false;
	}
	if (!varset) {
		LOGERROR("VariableSet has not been set");
		return false;
	}

	if (Initialized) {
		Flush();
		Initialized = false;
	}

	FILE* file;
	if (binary) {
		file = fopen(filename.c_str(), "wb");
	} else {
		file = fopen(filename.c_str(), "w");
	}

	if (!file) {
		LOGERROR("Failed to open output file \"%s\"", filename.c_str());
		return false;
	} else {
		if (!binary) {
			for (size_t i = 0; i < varset->GetNumVariables(); i++) {
				fprintf(file, "%s\t", varset->GetVariableName(i).c_str());
			}
			fprintf(file, "log prior\tlog likelihood\tlog posterior\tweight\n");
		}
		fclose(file);
	}

	output_buffer_ix = 0;
	Initialized = true;
	return true;
}

void PosteriorSampleHandlerFileDumper::Handle(const Real* values, Real lposterior, Real llh, Real lprior, Real weight, void*)
{
	ASSERT(Initialized);

	Sample& sample = output_buffer[output_buffer_ix];
	for (size_t i = 0; i < varset->GetNumVariables(); i++) {
		sample.values[i] = values[i];
	}
	sample.lprior = lprior;
	sample.llh = llh;
	sample.lposterior = lposterior;
	sample.weight = weight;

	output_buffer_ix++;
	if (output_buffer_ix == output_buffer.size()) {
		Flush();
	}
}

boost::function<void (const Real* values, Real lposterior, Real llh, Real lprior, Real weight, void* user)> PosteriorSampleHandlerFileDumper::GetFunctionPointer()
{
	ASSERT(Initialized);
	return boost::bind(&bcm::PosteriorSampleHandlerFileDumper::Handle, this, _1, _2, _3, _4, _5, _6);
}

void PosteriorSampleHandlerFileDumper::Flush()
{
	FILE* file;
	if (binary) {
		file = fopen(filename.c_str(), "ab");
	} else {
		file = fopen(filename.c_str(), "a");
	}
	if (file) {
		for (size_t j = 0; j < output_buffer_ix; j++) {
			const Sample& sample = output_buffer[j];
				
			if (binary) {
				float x;
				for (size_t i = 0; i < varset->GetNumVariables(); i++) {
					x = (float)sample.values[i];
					fwrite(&x, 4, 1, file);
				}
				x = (float)sample.lprior;
				fwrite(&x, 4, 1, file);
				x = (float)sample.llh;
				fwrite(&x, 4, 1, file);
				x = (float)sample.lposterior;
				fwrite(&x, 4, 1, file);
				x = (float)sample.weight;
				fwrite(&x, 4, 1, file);
			} else {
				for (size_t i = 0; i < varset->GetNumVariables(); i++) {
					fprintf(file, "%.6g\t", sample.values[i]);
				}
				fprintf(file, "%.6g\t", sample.lprior);
				fprintf(file, "%.6g\t", sample.llh);
				fprintf(file, "%.6g\t", sample.lposterior);
				fprintf(file, "%.6g\n", sample.weight);
			}
		}
		fclose(file);
	} else {
		LOGERROR("Failed to open output file \"%s\"", filename.c_str());
	}
	output_buffer_ix = 0;
}

}
