#include "Utils.h"
#include "PriorIndependence.h"
#include "UnivariateMarginal.h"
#include "VariableSet.h"

#include <boost/property_tree/xml_parser.hpp>

namespace bcm {

PriorIndependence::PriorIndependence(VariableSet* varset)
	: Prior(varset)
{
	Variables.resize(varset->GetNumVariables(), NULL);
}

PriorIndependence::~PriorIndependence()
{
	for (std::vector<UnivariateMarginal*>::iterator vi = Variables.begin(); vi != Variables.end(); ++vi) {
		bcm::safe_delete(*vi);
	}
}

bool PriorIndependence::LoadFromXML(const std::string& filename)
{
	boost::property_tree::ptree pt;
	try {
		boost::property_tree::read_xml(filename, pt);
	} catch (boost::property_tree::xml_parser_error &e) {
		LOGERROR("Error loading variable file: %s", e.what());
		return false;
	}
	
	try {
		boost::property_tree::ptree model_node = pt.get_child("variableset");
		
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, model_node.get_child("")) {
			if (var.first == "variable") {
				std::string name = var.second.get<std::string>("<xmlattr>.name");
				
				UnivariateMarginal* m = new UnivariateMarginal;
				if (!m->Initialize(var.second)) {
					return false;
				}
				if (!SetMarginal(name, m)) {
					return false;
				}
			}
		}

		return true;
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing variable file: %s", e.what());
		return false;
	}
}

bool PriorIndependence::SetMarginal(const std::string& name, UnivariateMarginal* marginal)
{
	size_t ix = varset->GetVariableIndex(name);
	if (ix == std::numeric_limits<size_t>::max()) {
		LOGERROR("Attempt to set marginal for unknown variable \"%s\"", name.c_str());
		return false;
	}

	Variables[ix] = marginal;
	LowerBounds[ix] = marginal->GetLowerBound();
	UpperBounds[ix] = marginal->GetUpperBound();
	return true;
}

bool PriorIndependence::EvaluateLogPDF(size_t threadix, const Real* values, Real& logp) const
{
	logp = 0.0;

	for (size_t i = 0; i < varset->GetNumVariables(); i++) {
		Real varlogp;
		if (!Variables[i]->EvaluateLogPDF(varlogp, values[i])) {
			LOG("Error evaluating prior for variable %d", i);
			return false;
		}
		if (varlogp != varlogp) {
			LOG("NaN in evaluating prior for variable %d, value=%g", i, values[i]);
		}
		logp += varlogp;
	}

	return true;
}

bool PriorIndependence::Sample(Real* values, RNG* rng) const
{
	for (size_t i = 0; i < varset->GetNumVariables(); i++) {
		if (!Variables[i]->Sample(values[i], rng)) {
			return false;
		}
	}
	return true;
}

bool PriorIndependence::EvaluateMarginalVariance(size_t i, Real& var) const
{
	if (i >= varset->GetNumVariables()) {
		return false;
	}
	return Variables[i]->EvaluateVariance(var);
}

bool PriorIndependence::EvaluateMarginalCDF(size_t i, const Real value, Real& p) const
{
	if (i >= varset->GetNumVariables()) {
		return false;
	}
	return Variables[i]->EvaluateCDF(p, value);
}

bool PriorIndependence::EvaluateMarginalQuantile(size_t i, const Real p, Real& value) const
{
	if (i >= varset->GetNumVariables()) {
		return false;
	}
	return Variables[i]->EvaluateQuantile(value, p);
}

}
