#include "Utils.h"
#include "Clustering.h"
#include "Prior.h"
#include "SamplerSMC.h"
#include "SMCSamplingStrategyBlockedMCMC.h"

namespace bcm {

SMCSamplingStrategyBlockedMCMC::SMCSamplingStrategyBlockedMCMC(SamplerSMC* sampler, TaskManager* task_manager)
	: SMCSamplingStrategy(sampler, task_manager)
	, recalculate_steps(20)
	, num_mcmc_steps_per_iteration(1)
	, update_step(0)
	, target_acceptance_rate(0.23)
{
	async.resize(num_threads);
	for (size_t k = 0; k < num_threads; k++) {
		async[k].proposed_var_values.resize(num_variables);
	}
}

SMCSamplingStrategyBlockedMCMC::~SMCSamplingStrategyBlockedMCMC()
{
}

void SMCSamplingStrategyBlockedMCMC::SetStepsPerIteration(size_t n)
{
	ASSERT(n > 0);
	num_mcmc_steps_per_iteration = n;
}

void SMCSamplingStrategyBlockedMCMC::SetRecalculateSteps(size_t n)
{
	ASSERT(n > 0);
	recalculate_steps = n;
}

bool SMCSamplingStrategyBlockedMCMC::UpdateStrategyIterationImpl(bool force_update, RNG* rng)
{
	if (force_update || update_step == recalculate_steps) {
		// Store previous scales
		previous_scale = VectorReal::Ones(num_variables);
		for (size_t i = 0; i < sampler_blocks.size(); i++) {
			Block& b = sampler_blocks[i];
			for (size_t vi = 0; vi < b.variable_indices.size(); vi++) {
				size_t ix = b.variable_indices[vi];
				previous_scale(ix) = b.scale;
			}
		}

		// Calculate correlation
		VectorReal empirical_mean = CalculatePreviousPopulationMean();
		MatrixReal empirical_covariance = CalculatePreviousPopulationCovariance(empirical_mean);
		MatrixReal empirical_correlation = empirical_covariance;
		for (size_t i = 0; i < num_variables; i++) {
			for (size_t j = 0; j < num_variables; j++) {
				if (j >= i) {
					empirical_correlation(i,j) = empirical_covariance(i,j) * bcm::rsqrt(empirical_covariance(i,i) * empirical_covariance(j,j));
				} else {
					empirical_correlation(i,j) = empirical_covariance(j,i) * bcm::rsqrt(empirical_covariance(i,i) * empirical_covariance(j,j));
				}
			}
		}
		
		// Calculate distance for clustering
		MatrixReal distance = MatrixReal::Zero(empirical_correlation.cols(), empirical_correlation.cols());
		for (size_t i = 0; i < (size_t)empirical_correlation.rows(); i++) {
			for (size_t j = 0; j < (size_t)empirical_correlation.cols(); j++) {
				distance(i,j) = 1.0 - fabs(empirical_correlation(i,j));
			}
		}

		// Hierarchical clustering and tree cutting
		std::vector< std::set<size_t> > clusters;
		bcm::TreeCluster(distance, clusters, 0.5);

		// Make parameter blocks
		sampler_blocks.clear();
		sampler_blocks.resize(clusters.size());
		for (size_t i = 0; i < clusters.size(); i++) {
			Block& b = sampler_blocks[i];

			for (std::set<size_t>::iterator vi = clusters[i].begin(); vi != clusters[i].end(); ++vi) {
				b.variable_indices.push_back(*vi);
			}

			if (clusters[i].size() > 1) {
				MatrixReal cov = MatrixReal::Zero(clusters[i].size(), clusters[i].size());
				for (size_t x = 0; x < clusters[i].size(); x++) {
					for (size_t y = 0; y < clusters[i].size(); y++) {
						cov(x, y) = empirical_covariance(b.variable_indices[x], b.variable_indices[y]);
					}
				}
				b.mvn_sampler.SetCovariance(cov);
			} else {
				b.sigma = VectorReal::Zero(clusters[i].size());
				for (size_t vi = 0; vi < clusters[i].size(); vi++) {
					size_t ix = b.variable_indices[vi];

					// Make sure the proposal variance is not too small
					Real var = empirical_covariance(ix,ix);
					Real prior_var;
					if (!prior->EvaluateMarginalVariance(ix, prior_var)) {
						// TODO - somehow come up with something reasonable?
						prior_var = 1.0;
					}
					var = std::max(var, 1e-6 * prior_var);
					b.sigma(vi) = sqrt(var);
				}
			}
			
			// Set the initial scale to the average scale that these parameters had in the previous round
			Real avg_scale = 0.0;
			for (size_t vi = 0; vi < b.variable_indices.size(); vi++) {
				size_t ix = b.variable_indices[vi];
				avg_scale += previous_scale(ix);
			}
			avg_scale /= (Real)b.variable_indices.size();
			b.scale = avg_scale;
		}
		for (size_t k = 0; k < num_threads; k++) {
			async[k].num_accepted.resize(sampler_blocks.size(), 0);
			async[k].num_proposed.resize(sampler_blocks.size(), 0);
		}

		LOG("Updated proposal:");
		for (size_t i = 0; i < sampler_blocks.size(); i++) {
			Block& b = sampler_blocks[i];
			std::string ixs = std::to_string((uint64)b.variable_indices[0]);
			for (size_t vi = 1; vi < b.variable_indices.size(); vi++) {
				ixs += std::string(",") + std::to_string((uint64)b.variable_indices[vi]);
			}
			LOG(" block %u: %g - %s", i, b.scale, ixs.c_str());
		}

		update_step = 1;
	} else {
		update_step++;
	}

	return true;
}

bool SMCSamplingStrategyBlockedMCMC::UpdateStrategyCurrentImpl(RNG* rng)
{
	for (size_t i = 0; i < sampler_blocks.size(); i++) {
		size_t num_accepted = 0;
		size_t num_proposed = 0;
		for (size_t k = 0; k < num_threads; k++) {
			num_accepted += async[k].num_accepted[i];
			num_proposed += async[k].num_proposed[i];
			async[k].num_accepted[i] = 0;
			async[k].num_proposed[i] = 0;
		}

		Real acceptance_rate = (Real)num_accepted / (Real)num_proposed;
		//LOG("Block %u -- Acceptance rate: %g - scale: %g", i, acceptance_rate, sampler_blocks[i].scale);
	
		// Adjust scaling to try to get the target acceptance rate.
		Real learn_rate = 1.0 + rng->GetReal() * 0.1;
		if (acceptance_rate < 0.91 * target_acceptance_rate) {
			sampler_blocks[i].scale /= learn_rate;
			sampler_blocks[i].scale = std::max(sampler_blocks[i].scale, (Real)1e-4);
		} else if (acceptance_rate > 1.1 * target_acceptance_rate) {
			sampler_blocks[i].scale *= learn_rate;
			sampler_blocks[i].scale = std::min(sampler_blocks[i].scale, (Real)1e2);
		}
	}

	return true;
}

bool SMCSamplingStrategyBlockedMCMC::UpdateParticleImpl(SamplerSMC::Particle& particle, size_t ix, RNG* rng, size_t threadix)
{
	bool result;

	for (size_t step = 0; step < num_mcmc_steps_per_iteration; step++) {
		for (size_t i = 0; i < sampler_blocks.size(); i++) {
			Block& b = sampler_blocks[i];
			
			async[threadix].proposed_var_values = particle.var_values;
			
			if (b.variable_indices.size() == 1) {
				// Scalar
				for (size_t vi = 0; vi < b.variable_indices.size(); vi++) {
					size_t ix = b.variable_indices[vi];
					Real x = rng->GetNormal(async[threadix].proposed_var_values(ix), b.scale * b.sigma(vi));
				
					// Reflect for variables with bounded support, this is still a symmetrical proposal
					const Real lb = prior->GetLowerBound(ix);
					const Real ub = prior->GetUpperBound(ix);
					while (1) {
						if (x < lb) {
							x = lb + (lb - x);
						} else if (x > ub) {
							x = ub - (x - ub);
						} else {
							break;
						}
					}

					async[threadix].proposed_var_values(ix) = x;
				}
			} else {
				// Covariance
				VectorReal x = VectorReal::Zero(b.variable_indices.size());
				b.mvn_sampler.Get(rng, x);
				
				for (size_t vi = 0; vi < b.variable_indices.size(); vi++) {
					size_t ix = b.variable_indices[vi];
					x[vi] *= b.scale;
					x[vi] += async[threadix].proposed_var_values(ix);
				
					// Reflect for variables with bounded support, this is still a symmetrical proposal
					const Real lb = prior->GetLowerBound(ix);
					const Real ub = prior->GetUpperBound(ix);
					while (1) {
						if (x[vi] < lb) {
							x[vi] = lb + (lb - x[vi]);
						} else if (x[vi] > ub) {
							x[vi] = ub - (x[vi] - ub);
						} else {
							break;
						}
					}
					
					async[threadix].proposed_var_values(ix) = x(vi);
				}
			}

			// Evaluate prior & likelihood
			Real new_lprior;
			Real new_llh;
			Real new_lpowerposterior;
			result = EvaluatePriorLikelihood(threadix, async[threadix].proposed_var_values.data(), new_lprior, new_llh);
			if (!result) {
				return false;
			}
			new_lpowerposterior = new_lprior + current_temperature * new_llh;

			// Decide on transition
			Real transition_probability = new_lpowerposterior - particle.lpowerposterior;
			transition_probability = exp(transition_probability);
			transition_probability = std::min(transition_probability, (Real)1.0);

			Real alpha = rng->GetReal();
			async[threadix].num_proposed[i]++;
			if (alpha < transition_probability) {
				async[threadix].num_accepted[i]++;

				particle.var_values = async[threadix].proposed_var_values;
				if (particle.output_values.size() > 0) {
					if (!GetOutputValues(threadix, particle.output_values.data())) {
						return false;
					}
				}
				particle.lprior = new_lprior;
				particle.llh = new_llh;
				particle.lpowerposterior = new_lpowerposterior;
			} else {
				// Revert to previous state
			}
		}
	}
	
	return true;
}

bool SMCSamplingStrategyBlockedMCMC::CalculateParticleWeightImpl(SamplerSMC::Particle& particle, size_t ix, size_t threadix)
{
	const SamplerSMC::Particle& prev_particle = (*previous_population)[ix];

	// Calculate new weight with approximate backward kernel
	if (prev_particle.normweight > 0) {
		Real incremental_weight = (prev_particle.lprior + current_temperature * prev_particle.llh) - prev_particle.lpowerposterior;
		particle.lweight = incremental_weight;
		particle.accumulated_lweight = prev_particle.accumulated_lweight + incremental_weight;
	} else {
		particle.lweight = -std::numeric_limits<Real>::infinity();
		particle.accumulated_lweight = -std::numeric_limits<Real>::infinity();
	}

	return true;
}

}
