#include "Utils.h"
#include "Clustering.h"
#include "Prior.h"
#include "SMCSamplingStrategyClusteredMCMC.h"

static const size_t MAX_CLUSTERS = 5;
static const size_t NUM_CLUSTER_PASSES = 3;

namespace bcm {

SMCSamplingStrategyClusteredMCMC::SMCSamplingStrategyClusteredMCMC(SamplerSMC* sampler, TaskManager* task_manager)
	: SMCSamplingStrategy(sampler, task_manager)
	, recalculate_steps(100)
	, num_mcmc_steps_per_iteration(1)
	, target_acceptance_rate(0.23)
	, update_step(0)
	, use_k(0)
{
	// Clustering
	clusters.resize(MAX_CLUSTERS);
	for (size_t i = 0; i < MAX_CLUSTERS; i++) {
		clusters[i].resize(i+1);
		for (size_t j = 0; j < i+1; j++) {
			clusters[i][j].empirical_mean = VectorReal::Zero(num_variables);
			clusters[i][j].empirical_variance = VectorReal::Zero(num_variables);
			clusters[i][j].proposal_sigma = VectorReal::Zero(num_variables);
			clusters[i][j].proposal_scale = 1.0;
		}
	}
	cluster_assignments.resize(MAX_CLUSTERS-1);
	for (size_t i = 0; i < MAX_CLUSTERS-1; i++) {
		cluster_assignments[i].resize(population_size);
	}
	clustering_bic.resize(MAX_CLUSTERS);

	// Asyncrhonous operation
	async.resize(num_threads);
	for (size_t k = 0; k < num_threads; k++) {
		async[k].proposed_var_values.resize(num_variables);
		async[k].num_accepted.resize(MAX_CLUSTERS, 0);
		async[k].num_proposed.resize(MAX_CLUSTERS, 0);
	}
}

SMCSamplingStrategyClusteredMCMC::~SMCSamplingStrategyClusteredMCMC()
{
}

void SMCSamplingStrategyClusteredMCMC::SetStepsPerIteration(size_t n)
{
	ASSERT(n > 0);
	num_mcmc_steps_per_iteration = n;
}

void SMCSamplingStrategyClusteredMCMC::SetRecalculateSteps(size_t n)
{
	ASSERT(n > 0);
	recalculate_steps = n;
}

bool SMCSamplingStrategyClusteredMCMC::UpdateStrategyIterationImpl(bool force_update, RNG* rng)
{
	if (force_update || update_step == recalculate_steps) {
		std::vector<uint64> tasks;
		tasks.resize(MAX_CLUSTERS-1);
		for (size_t i = 0; i < tasks.size(); i++) {
			TTask task = boost::bind(&SMCSamplingStrategyClusteredMCMC::ClusterSamples, this, _1, _2);
			tasks[i] = task_manager->AddTask(task, reinterpret_cast<void*>(i));
		}
		for (size_t i = 0; i < tasks.size(); i++) {
			task_manager->WaitTask(tasks[i]);
		}

		// Also calculate unclustered BIC
		CalculateBIC(1);

		LOG("Clustering statistics:");
		for (size_t k = 1; k <= MAX_CLUSTERS; k++) {
			LOG("  k=%d: BIC=%.5g", k, clustering_bic[k-1]);
		}
		for (size_t k = 1; k <= MAX_CLUSTERS; k++) {
			LOG("  k=%d:", k, clustering_bic[k-1]);
			for (size_t j = 0; j < k; j++) {
				LOG("    cluster %d - parameter 1 mean=%g", j, clusters[k-1][j].empirical_mean(0));
			}
		}

		// Select k
		use_k = 1;
		Real lowest_bic = clustering_bic[0];
		for (size_t k = 2; k <= MAX_CLUSTERS; k++) {
			if (clustering_bic[k-1] < lowest_bic) {
				use_k = k;
				lowest_bic = clustering_bic[k-1];
			}
		}
		LOG("Selected k=%d", use_k);
		
		for (size_t j = 0; j < use_k; j++) {
			// Set proposal variance for the next iteration
			Cluster& c = clusters[use_k-1][j];
			for (size_t i = 0; i < num_variables; i++) {
				Real prior_var;
				if (!prior->EvaluateMarginalVariance(i, prior_var)) {
					// TODO - somehow come up with something reasonable?
					prior_var = 1.0;
				}
				Real proposal_var = std::max(c.empirical_variance(i), 1e-6 * prior_var);
				c.proposal_sigma(i) = sqrt(proposal_var);
			}
		}

		update_step = 1;
	} else {
		update_step++;
	}

	return true;
}

bool SMCSamplingStrategyClusteredMCMC::UpdateStrategyCurrentImpl(RNG* rng)
{
	// Update proposal scaling
	for (size_t i = 0; i < use_k; i++) {
		Cluster& c = clusters[use_k-1][i];

		size_t num_accepted = 0;
		size_t num_proposed = 0;
		for (size_t k = 0; k < num_threads; k++) {
			num_accepted += async[k].num_accepted[i];
			num_proposed += async[k].num_proposed[i];
			async[k].num_accepted[i] = 0;
			async[k].num_proposed[i] = 0;
		}

		Real acceptance_rate = (Real)num_accepted / (Real)num_proposed;
		//LOG("Acceptance rate: %g - scale: %g", acceptance_rate, proposal_scale);
	
		// Adjust scaling to try to get the target acceptance rate.
		Real learn_rate = 1.0 + rng->GetReal() * 0.1;
		if (acceptance_rate < 0.91 * target_acceptance_rate) {
			c.proposal_scale /= learn_rate;
			c.proposal_scale = std::max(c.proposal_scale, (Real)1e-10);
		} else if (acceptance_rate > 1.1 * target_acceptance_rate) {
			c.proposal_scale *= learn_rate;
			c.proposal_scale = std::min(c.proposal_scale, (Real)1e3);
		}
	}

	return true;
}

bool SMCSamplingStrategyClusteredMCMC::UpdateParticleImpl(SamplerSMC::Particle& particle, size_t ix, RNG* rng, size_t threadix)
{
	bool result;

	// Find cluster assignment
	size_t cluster_ix;
	if (use_k == 1) {
		cluster_ix = 0;
	} else {
		cluster_ix = cluster_assignments[use_k-2][ix];
	}
	const Cluster& c = clusters[use_k-1][cluster_ix];
	
	for (size_t step = 0; step < num_mcmc_steps_per_iteration; step++) {
		// Propose new parameters
		for (size_t i = 0; i < num_variables; i++) {
			Real x = rng->GetNormal(0.0, c.proposal_scale * c.proposal_sigma(i));
			Real nx = particle.var_values[i] + x;
			async[threadix].proposed_var_values(i) = nx;
		}
		
		// Reflect for variables with bounded support, this is still a symmetrical proposal
		for (size_t i = 0; i < num_variables; i++) {
			const Real lb = prior->GetLowerBound(i);
			const Real ub = prior->GetUpperBound(i);
			Real nx = async[threadix].proposed_var_values(i);
			while (1) {
				if (nx < lb) {
					nx = lb + (lb - nx);
				} else if (nx > ub) {
					nx = ub - (nx - ub);
				} else {
					break;
				}
			}
			async[threadix].proposed_var_values(i) = nx;
		}

		// Evaluate prior & likelihood
		Real new_lprior;
		Real new_llh;
		Real new_lpowerposterior;
		result = EvaluatePriorLikelihood(threadix, async[threadix].proposed_var_values.data(), new_lprior, new_llh);
		if (!result) {
			return false;
		}
		new_lpowerposterior = new_lprior + current_temperature * new_llh;

		// Decide on transition
		Real transition_probability = new_lpowerposterior - particle.lpowerposterior;
		transition_probability = exp(transition_probability);
		transition_probability = std::min(transition_probability, (Real)1.0);

		Real alpha = rng->GetReal();
		async[threadix].num_proposed[cluster_ix]++;
		if (alpha < transition_probability) {
			async[threadix].num_accepted[cluster_ix]++;

			particle.var_values = async[threadix].proposed_var_values;
			if (particle.output_values.size() > 0) {
				if (!GetOutputValues(threadix, particle.output_values.data())) {
					return false;
				}
			}
			particle.lprior = new_lprior;
			particle.llh = new_llh;
			particle.lpowerposterior = new_lpowerposterior;
		} else {
			// Revert to previous state
		}
	}
	
	return true;
}

bool SMCSamplingStrategyClusteredMCMC::CalculateParticleWeightImpl(SamplerSMC::Particle& particle, size_t ix, size_t threadix)
{
	const SamplerSMC::Particle& prev_particle = (*previous_population)[ix];

	// Calculate new weight with approximate backward kernel
	if (prev_particle.normweight > 0) {
		Real incremental_weight = (prev_particle.lprior + current_temperature * prev_particle.llh) - prev_particle.lpowerposterior;
		particle.lweight = incremental_weight;
		particle.accumulated_lweight = prev_particle.accumulated_lweight + incremental_weight;
	} else {
		particle.lweight = -std::numeric_limits<Real>::infinity();
		particle.accumulated_lweight = -std::numeric_limits<Real>::infinity();
	}

	return true;
}

void SMCSamplingStrategyClusteredMCMC::ClusterSamples(void* user, size_t thread)
{
	size_t task = reinterpret_cast<size_t>(user);
	int k = (int)task+2;

	Real error;
	std::vector<const Real*> data(population_size);
	std::vector<Real> weights(population_size);
	std::vector<int> assignment(population_size, -1);
	for (size_t i = 0; i < population_size; i++) {
		data[i] = (*previous_population)[i].var_values.data();
		weights[i] = (*previous_population)[i].normweight * population_size;
	}

	KCluster(k, population_size, num_variables, data.data(), weights.data(), NUM_CLUSTER_PASSES, &assignment[0], &error);
	
	char fn[20];
	sprintf(fn, "samples_%d.tsv", k);
	FILE* file = fopen(fn, "w");
	for (size_t i = 0; i < population_size; i++) {
		(cluster_assignments[task])[i] = (size_t)assignment[i];
		fprintf(file, "%g\t%g\t%d\n", (*previous_population)[i].var_values[0], (*previous_population)[i].var_values[1], assignment[i]);
	}
	fclose(file);

	CalculateBIC(k);
}

void SMCSamplingStrategyClusteredMCMC::CalculateBIC(size_t k)
{
	Real llh = 0.0;
	for (size_t i = 0; i < k; i++) {
		// Calculate empirical mean
		VectorReal& mean = clusters[k-1][i].empirical_mean;
		mean.setZero();
		Real acc_weight = 0.0;
		for (size_t j = 0; j < population_size; j++) {
			if (k == 1 || cluster_assignments[k-2][j] == i) {
				const VectorReal& s = (*previous_population)[j].var_values;
				const Real w = (*previous_population)[j].normweight * previous_population->size();
				if (w > 0.0) {
					Real one_over_kp1 = w / (acc_weight + w);
					for (size_t l = 0; l < num_variables; l++) {
						mean(l) += (s(l) - mean(l)) * one_over_kp1;
					}
					acc_weight += w;
				}
			}
		}
		
		// Calculate empirical variances
		VectorReal& variance = clusters[k-1][i].empirical_variance;
		variance.setZero();
		acc_weight = 0.0;
		for (size_t j = 0; j < population_size; j++) {
			if (k == 1 || cluster_assignments[k-2][j] == i) {
				const VectorReal& s = (*previous_population)[j].var_values;
				const Real w = (*previous_population)[j].normweight * previous_population->size();
				if (w > 0.0) {
					Real one_over_kp1 = w / (acc_weight + w);
					for (size_t l = 0; l < num_variables; l++) {
						Real d = s(l) - mean(l);
						variance(l) += (d * d - variance(l)) * one_over_kp1;
					}
					acc_weight += w;
				}
			}
		}

		// Calculate likelihood
		acc_weight = 0.0;
		for (size_t j = 0; j < population_size; j++) {
			if (k == 1 || cluster_assignments[k-2][j] == i) {
				const VectorReal& s = (*previous_population)[j].var_values;
				const Real w = (*previous_population)[j].normweight * previous_population->size();
				if (w > 0.0) {
					Real one_over_kp1 = w / (acc_weight + w);
					for (size_t l = 0; l < num_variables; l++) {
						Real d = s(l) - mean(l);
						llh -= ((d * d) / (2.0 * variance(l))) * one_over_kp1;
					}
					acc_weight += w;
				}
			}
		}
		
		for (size_t l = 0; l < num_variables; l++) {
			llh += acc_weight * log(bcm::rsqrt(2.0 * M_PI * variance(l)));
		}
	}
	
	clustering_bic[k-1] = -2.0 * llh + (2.0 * num_variables * k) * log((Real)population_size);
}

}
