#include "Utils.h"
#include "Likelihood.h"
#include "Prior.h"
#include "ProgressIndicator.h"
#include "RNG.h"
#include "Sampler.h"
#include "SamplerFactory.h"
#include "VariableSet.h"

#include <boost/filesystem.hpp>

namespace bcm {

Sampler::Sampler(size_t num_threads)
	: have_reentrant_prior(false)
	, have_reentrant_likelihood(false)
	, prior(NULL)
	, likelihood(NULL)
	, varset(NULL)
	, rngseed(0)
	, num_variables(0)
	, num_output_values(0)
	, num_threads(num_threads)
	, posterior_sample_handler_user(NULL)
	, output_values_handler_user(NULL)
	, evidence_handler_user(NULL)
	, progress(NULL)
	, runtime_seconds(0)
	, multistart_continue(false)
	, initialized(false)
	, num_likelihood_evaluations(0)
	, evidence(-std::numeric_limits<Real>::infinity())
	, evidence_sd(-std::numeric_limits<Real>::infinity())
	, task_manager(new bcm::TaskManager(num_threads))
{
	num_parallel_likelihood_evaluations.resize(num_threads, 0);
}

Sampler::~Sampler()
{
	delete task_manager;
}

void Sampler::SetLatentVariableSet(VariableSet* varset)
{
	this->varset = varset;
	num_variables = varset->GetNumVariables();
}

void Sampler::SetReentrantPrior(Prior* prior)
{
	have_reentrant_prior = true;
	this->prior = prior;
}

void Sampler::SetReentrantLikelihood(Likelihood* likelihood)
{
	have_reentrant_likelihood = true;
	this->likelihood = likelihood;
}

bool Sampler::SetNonreentrantPrior(size_t i, Prior* prior)
{
	have_reentrant_prior = false;
	if (i == 0) {
		// Store the first index also in the regular prior pointer, to be used by the non-parallelized code.
		this->prior = prior;
		nonreentrant_priors[i] = prior;
	} else if (i < num_threads) {
		nonreentrant_priors[i] = prior;
	} else {
		LOGERROR("Non-reentrant prior given for thread %u, but sampler was constructed for only %u threads", i, num_threads);
		return false;
	}
	return true;
}

bool Sampler::SetNonreentrantLikelihood(size_t i, Likelihood* likelihood)
{
	have_reentrant_likelihood = false;
	if (i == 0) {
		// Store the first index also in the regular likelihood pointer, to be used by the non-parallelized code.
		this->likelihood = likelihood;
		nonreentrant_likelihoods[i] = likelihood;
	} else if (i < num_threads) {
		nonreentrant_likelihoods[i] = likelihood;
	} else {
		LOGERROR("Non-reentrant likelihood given for thread %u, but sampler was constructed for only %u threads", i, num_threads);
		return false;
	}
	return true;
}

void Sampler::SetOutputPath(const std::string path)
{
	output_path = path;
	bcm::fix_path(output_path);
}

void Sampler::SetPosteriorSampleHandler(TPosteriorSampleHandler fp, void* user)
{
	posterior_sample_handler = fp;
	posterior_sample_handler_user = user;
}

void Sampler::SetOutputValuesHandler(TOutputValuesHandler fp, void* user)
{
	output_values_handler = fp;
	output_values_handler_user = user;
}

void Sampler::SetEvidenceHandler(TEvidenceHandler fp, void* user)
{
	evidence_handler = fp;
	evidence_handler_user = user;
}

void Sampler::SetProgressIndicator(ProgressIndicator* pi)
{
	progress = pi;
}

bool Sampler::Parse(const boost::program_options::variables_map& vm)
{
	if (initialized) {
		LOGERROR("Sampler is already initialized");
		return false;
	}
	
	try {
		rngseed = vm["sampler.rngseed"].as<unsigned long>();
		//runtime_seconds	= vm["runtime_seconds"].as<size_t>();
		multiple_start_state_file = vm["sampler.multistart_state_file"].as<std::string>();
		if (vm.count("continue")) {
			multistart_continue = true;
		} else {
			multistart_continue = false;
		}
	} catch (boost::program_options::error& e) {
		LOGERROR("Error parsing sampler: %s", e.what());
		return false;
	}

	return ParseImpl(vm);
}

bool Sampler::Initialize()
{
	if (initialized) {
		LOGERROR("Sampler is already initialized");
		return false;
	}

	if (rngseed == 0) {
		rngseed = RNG::GetTimeBasedSeed();
	}
	LOG("Using random number seed %u", rngseed);

	if (have_reentrant_prior) {
		if (!prior) {
			LOGERROR("No prior specified");
			return false;
		}
	} else {
		bool have_prior = (prior == NULL) ? false : true;
		for (size_t i = 0; i < num_threads; i++) {
			if (!nonreentrant_priors[i]) {
				have_prior = false;
				break;
			}
		}
		if (!have_prior) {
			LOGERROR("Prior has not been specified for all threads");
			return false;
		}
	}

	if (have_reentrant_likelihood) {
		if (!likelihood) {
			LOGERROR("No likelihood specified");
			return false;
		}
	} else {
		bool have_likelihood = (likelihood == NULL) ? false : true;
		for (size_t i = 0; i < num_threads; i++) {
			if (!nonreentrant_likelihoods[i]) {
				have_likelihood = false;
				break;
			}
		}
		if (!have_likelihood) {
			LOGERROR("Likelihood has not been specified for all threads");
			return false;
		}
	}

	if (!varset) {
		LOGERROR("No variable set specified");
		return false;
	}
	if (num_variables == 0) {
		LOGERROR("No variables to sample");
		return false;
	}

	if (!output_path.empty()) {
		if (!boost::filesystem::is_directory(output_path)) {
			std::string create_path = output_path.substr(0, output_path.size()-1);
			if (!boost::filesystem::create_directories(create_path)) {
				LOGERROR("Could not create output directory.");
				return false;
			}
		}
	}

	if (!likelihood->GetOutput(0, NULL, num_output_values)) {
		LOGERROR("Query for number of output values failed, assuming 0 output values");
	}

	if (!InitializeImpl()) {
		return false;
	}

	initialized = true;
	return true;
}

void Sampler::Uninitialize()
{
	if (!initialized) {
		LOGERROR("Sampler is not initialized");
		return;
	}

	UninitializeImpl();

	initialized = false;
}

bool Sampler::Run(Real* levidence, Real* levidence_sd)
{
	if (!initialized) {
		LOGERROR("Sampler has not been initialized");
		return false;
	}

	// Check if we're continuing from a previous state.
	if (multistart_continue) {
		if (!LoadCurrentState()) {
			return false;
		}
	} else {
		evidence = std::numeric_limits<Real>::quiet_NaN();
		evidence_sd = std::numeric_limits<Real>::quiet_NaN();
	}

	LOG("Starting integration...");

	bcm::Timer timer;
	timer.Start();
	num_likelihood_evaluations = 0;
	for (size_t i = 0; i < num_threads; i++) {
		num_parallel_likelihood_evaluations[i] = 0;
	}
	if (progress) {
		progress->NotifyStart();
	}

	ERunResult result = RunImpl();

	if (progress) {
		progress->NotifyStop();
	}
	double integration_time = timer.GetElapsedSeconds();
	for (size_t i = 0; i < num_threads; i++) {
		num_likelihood_evaluations += num_parallel_likelihood_evaluations[i];
	}
	
	LOG("");
	LOG("Timing information:");
	LOG("  Integration time: %.3g seconds", integration_time);
	LOG("  Number of likelihood evaluations: %u", num_likelihood_evaluations);
	LOG("  Evaluations per second: %.0f", num_likelihood_evaluations / integration_time);
	LOG("");
	LOG("CPU information:");
	bcm::logger->LogCPUInfo();
	
	if (have_reentrant_likelihood) {
		likelihood->PostInference(output_path);
	} else {
		for (size_t i = 0; i < num_threads; i++) {
			nonreentrant_likelihoods[i]->PostInference(output_path);
		}
	}
	
	if (result == RUN_FINISHED) {
		LOG("");
		LOG("Sampling finished.");
		LOG("Evidence: %g +/- %g", evidence, evidence_sd);

		if (levidence) {
			*levidence = evidence;
		}
		if (levidence_sd) {
			*levidence_sd = evidence_sd;
		}
		return true;
	} else if (result == RUN_TIMED_END) {
		LOG("Alotted time has been used, saving current sampler state in %s", multiple_start_state_file.c_str());
		if (!SaveCurrentState()) {
			return false;
		} else {
			return true;
		}
	} else if (result == RUN_ERROR) {
		LOGERROR("Sampling failed");
		return false;
	} else {
		return false;
	}
}

bool Sampler::EvaluatePrior(size_t threadix, const Real* values, Real& lprior)
{
	lprior = -std::numeric_limits<Real>::infinity();
	if (threadix >= num_threads) {
		LOGERROR("Invalid thread index %u", (unsigned int)threadix);
		return false;
	}
	
	bool result;
	if (have_reentrant_prior) {
		result = prior->EvaluateLogPDF(threadix, values, lprior);
	} else {
		result = nonreentrant_priors[threadix]->EvaluateLogPDF(threadix, values, lprior);
	}
	if (!result) {
		LOGERROR("Prior evaluation failed");
	} else if (lprior != lprior) {
		LOGERROR("NAN in prior calculation");
		result = false;
	}
	return result;
}

bool Sampler::EvaluateLikelihood(size_t threadix, const Real* values, Real& llh)
{
	llh = -std::numeric_limits<Real>::infinity();
	if (threadix >= num_threads) {
		LOGERROR("Invalid thread index %u", (unsigned int)threadix);
		return false;
	}
	
	bool result;
	if (have_reentrant_likelihood) {
		result = likelihood->EvaluateLogProbability(threadix, values, llh);
	} else {
		result = nonreentrant_likelihoods[threadix]->EvaluateLogProbability(threadix, values, llh);
	}
	num_parallel_likelihood_evaluations[threadix]++;
	if (!result) {
		LOGERROR("Likelihood evaluation failed");
	} else if (llh != llh) {
		LOGERROR("NAN in likelihood calculation with values:");
		for (size_t i = 0; i < num_variables; i++) {
			LOGERROR("  %s: %g", varset->GetVariableName(i).c_str(), values[i]);
		}
		result = false;
	}
	return result;
}

bool Sampler::EvaluatePriorLikelihood(size_t threadix, const Real* values, Real& lprior, Real& llh)
{
	if (!EvaluatePrior(threadix, values, lprior)) {
		return false;
	}
	if (lprior == -std::numeric_limits<Real>::infinity()) {
		llh = -std::numeric_limits<Real>::infinity();
		return true;
	} else {
		return EvaluateLikelihood(threadix, values, llh);
	}
}

bool Sampler::GetOutputValues(size_t threadix, Real* values)
{
	bool result = true;
	if (num_output_values > 0) {
		if (have_reentrant_likelihood) {
			result = likelihood->GetOutput(threadix, values, num_output_values);
		} else {
			result = nonreentrant_likelihoods[threadix]->GetOutput(threadix, values, num_output_values);
		}
	}
	if (!result) {
		LOGERROR("Retrieving output values from likelihood failed");
	}
	return result;
}

bool Sampler::SaveCurrentState()
{
	std::ofstream ofs;
	ofs.open(output_path + multiple_start_state_file, std::ofstream::binary);
	if (!ofs.is_open()) {
		LOGERROR("Failed to open file \"%s\" for saving sampler state", multiple_start_state_file.c_str());
		return false;
	}

	{
		boost::archive::binary_oarchive ar(ofs);
		SamplerFactory::SerializeSampler(this, ar);
	}

	ofs.close();

	return true;
}

bool Sampler::LoadCurrentState()
{
	std::ifstream ifs;
	ifs.open(output_path + multiple_start_state_file, std::ofstream::binary);
	if (!ifs.is_open()) {
		LOGERROR("Failed to open file \"%s\" for loading sampler state", multiple_start_state_file.c_str());
		return false;
	}

	{
		boost::archive::binary_iarchive ar(ifs);
		SamplerFactory::SerializeSampler(this, ar);
	}

	ifs.close();

	return true;
}

}
