#include "Utils.h"
#include "CSVParser.h"
#include "ProgressIndicator.h"
#include "RNG_SFMT.h"
#include "SamplerPT.h"

namespace bcm {

SamplerPT::SamplerPT(size_t num_threads)
	: Sampler(num_threads)
	, num_chains(16)
	, burnin_period(1000)
	, sampling_period(4000)
	, use_every_nth(10)
	, num_mcmc_moves_per_iteration(1)
	, min_temperature(0.0)
	, initial_temperature_power(5.0)
	, exchange_probability(0.5)
	, adapt_proposal_samples(2000)
	, proposal_scaling_learning_rate(0.005)
	, proposal_scaling_ema_period(1000)
	, target_acceptance_rate(0.23)
	, adapt_temperatures_times(0)
	, adapt_temperatures_period(1000)
	, proposal_type(PT_FullVector)
	, rng(NULL)
{
}

SamplerPT::~SamplerPT()
{
	bcm::safe_delete(rng);
}

bool SamplerPT::ParseImpl(const boost::program_options::variables_map& vm)
{
	try {
		burnin_period = vm["sampler.burnin_period"].as<size_t>();
		sampling_period = vm["sampler.sampling_period"].as<size_t>();
		use_every_nth = vm["sampler.use_every_nth"].as<size_t>();
		num_mcmc_moves_per_iteration = vm["sampler.num_mcmc_moves_per_iteration"].as<size_t>();
		num_chains = vm["sampler.numchains"].as<size_t>();
		min_temperature = vm["sampler.min_temperature"].as<Real>();
		temperatures_file = vm["sampler.temperatures"].as<std::string>();
		initial_temperature_power = vm["sampler.initial_temperature_power"].as<Real>();
		adapt_proposal_samples = vm["sampler.adapt_proposal"].as<size_t>();
		exchange_probability = vm["sampler.exchange_probability"].as<Real>();
		adapt_temperatures_times = vm["sampler.adapt_temperatures_times"].as<size_t>();
		adapt_temperatures_period = vm["sampler.adapt_temperatures_period"].as<size_t>();

		proposal_scaling_ema_period = (size_t)ceil(adapt_proposal_samples * use_every_nth * num_mcmc_moves_per_iteration * (1.0 - exchange_probability) / 10);
		proposal_scaling_learning_rate = pow(100.0, 1.0 / proposal_scaling_ema_period) - 1.0;

		if (vm["sampler.proposal_type"].as<std::string>() == "scalar") {
			proposal_type = PT_Scalar;
		} else if (vm["sampler.proposal_type"].as<std::string>() == "vector") {
			proposal_type = PT_FullVector;
		} else if (vm["sampler.proposal_type"].as<std::string>() == "autoblock") {
			proposal_type = PT_AutoBlock;
		} else {
			LOGERROR("Proposal type must be either: scalar, vector, autoblock");
			return false;
		}

	} catch (boost::program_options::error& e) {
		LOGERROR("Error parsing sampler: %s", e.what());
		return false;
	}

	return true;
}

bool SamplerPT::InitializeImpl()
{
	rng = new RNG_SFMT(rngseed);
	
	temperatures.resize(num_chains);
	if (!temperatures_file.empty()) {
		// Read temperatures from the specified file.
		CSVParser parser;
		if (!parser.Parse(temperatures_file, "\t", false)) {
			return false;
		}

		if (parser.GetNumColumns() != 1) {
			LOGERROR("Temperature file \"%s\" has incorrect number of columns (%u)", temperatures_file.c_str(), (unsigned int)parser.GetNumColumns());
			return false;
		}
		if (parser.GetNumRows() != num_chains) {
			LOGERROR("Temperature file \"%s\" has incorrect number of rows (%u), expected %u", temperatures_file.c_str(), (unsigned int)parser.GetNumRows(), num_chains);
			return false;
		}
		
		for (size_t i = 0; i < num_chains; i++) {
			temperatures(i) = parser.GetEntry(i, 0);
		}
	} else {
		// Calculate initial temperatures using a log schedule
		temperatures[0] = min_temperature;
		temperatures[num_chains-1] = 1.0;

		if (min_temperature == 0.0) {
			// Power-law scaling from 1.0 downwards
			for (size_t i = 1; i < num_chains-1; i++) {
				Real alpha = i / (Real)(num_chains-1);
				temperatures[i] = pow(alpha, initial_temperature_power);
			}
		} else {
			// Log-linear from min to 1.0.
			Real alpha = pow(min_temperature, 1.0 / (Real)(num_chains - 1));
			alpha = std::max(alpha, std::numeric_limits<Real>::epsilon());
			for (size_t i = num_chains - 2; i > 0; i--) {
				temperatures[i] = alpha * temperatures[i+1];
			}
		}
	}

	// Initialize chains
	chains.resize(num_chains, SamplerPTChain(this));
	for (size_t i = 0; i < num_chains; i++) {
		chains[i].SetTemperature(temperatures[i]);
		chains[i].SetMCMCMovesPerIteration(num_mcmc_moves_per_iteration);
		bool result = chains[i].Initialize(rng->GetInt());
		if (!result) {
			return false;
		}
	}
	samples_lposterior.resize(num_chains, VectorReal::Zero(sampling_period));

	// Thermodynamic integration
	thermodynamic_integrand = VectorReal::Zero(num_chains);
	thermodynamic_integrand_variance = VectorReal::Zero(num_chains);

	// Feedback optimization
	up_count = VectorReal::Zero(num_chains);
	down_count = VectorReal::Zero(num_chains);
	labels = Eigen::VectorXi::Zero(num_chains);
	up_labeled_time = VectorReal::Zero(num_chains);
	down_labeled_time = VectorReal::Zero(num_chains);

	return true;
}

void SamplerPT::UninitializeImpl()
{
	chains.clear();
}

Sampler::ERunResult SamplerPT::RunImpl()
{
	LOG("Finding starting point...");
	for (size_t i = 0; i < num_chains; i++) {
		bool result = chains[i].FindStartingPosition();
		if (!result) {
			return RUN_ERROR;
		}
	}

	Real max_lposterior = -std::numeric_limits<Real>::infinity();
	
	bool result = true;
	if (adapt_temperatures_times > 0) {
		for (size_t ati = 0; ati < adapt_temperatures_times; ati++) {
			LOG("Starting temperature adaptation run %u", (unsigned int)ati+1);
			
			size_t totalSamples = adapt_temperatures_period * use_every_nth;
			for (size_t si = 0; si < totalSamples; si++) {
				if (si % (totalSamples / 100) == 0) {
					double progress = 100.0 * si / (double)totalSamples;
					LOG("Progress %.0f%%", progress);
				}
				if (si % (std::min(use_every_nth, (size_t)100)) == 0) {
					UpdateProgress((ati * adapt_temperatures_period * use_every_nth + si) / (Real)(adapt_temperatures_times * adapt_temperatures_period * use_every_nth + (burnin_period + sampling_period) * use_every_nth));
				}

				if (si == totalSamples / 2) {
					LOG("Updating proposal to empirical variance");
					for (size_t i = 0; i < num_chains; i++) {
						chains[i].AdaptProposal();
					}
				}

				if (num_chains > 1) {
					// Randomly choose between mutate and exchange moves
					Real exchange = rng->GetReal();
					if (exchange < exchange_probability) {
						DoExchangeMove(true, si);
					} else {
						result = DoMutateMove(true);
					}
				} else {
					result = DoMutateMove(true);
				}

				if (!result) {
					LOGERROR("Sample step failed");
					LogStatistics();
					return RUN_ERROR;
				}

				if (chains[num_chains-1].GetLogPowerPosterior() > max_lposterior) {
					max_lposterior = chains[num_chains-1].GetLogPowerPosterior();
					LOG("New max log posterior: %g", max_lposterior);
				}
			}

			LogStatistics();
			AdaptTemperatures();
		}
	}

	LOG("Starting sampling loop...");
	size_t totalSamples = (burnin_period + sampling_period) * use_every_nth;
	for (size_t si = 0; si < totalSamples; si++) {
		if (totalSamples > 100) {
			if (si % (totalSamples / 100) == 0) {
				double progress = 100.0 * si / (Real)totalSamples;
				LOG("Progress %.0f%%", progress);
			}
			if (si % (std::min(use_every_nth, (size_t)100)) == 0) {
				if (adapt_temperatures_times > 0) {
					UpdateProgress((adapt_temperatures_times * adapt_temperatures_period * use_every_nth + si) / (Real)(adapt_temperatures_times * adapt_temperatures_period * use_every_nth + totalSamples));
				} else {
					UpdateProgress(si / (Real)totalSamples);
				}
			}
		} else {
			double progress = 100.0 * si / (Real)totalSamples;
			LOG("Progress %.0f%%", progress);
			UpdateProgress(si / (Real)totalSamples);
		}

		size_t sampleIx = si / use_every_nth;
		bool burnin = sampleIx < burnin_period;

		if (num_chains > 1) {
			// Randomly choose between mutate and exchange moves
			Real exchange = rng->GetReal();
			if (exchange < exchange_probability) {
				DoExchangeMove(burnin, si);
			} else {
				result = DoMutateMove(burnin);
			}
		} else {
			result = DoMutateMove(burnin);
		}

		if (!result) {
			LOGERROR("Sample step failed");
			LogStatistics();
			return RUN_ERROR;
		}

		if (chains[num_chains-1].GetLogPowerPosterior() > max_lposterior) {
			max_lposterior = chains[num_chains-1].GetLogPowerPosterior();
			LOG("New max log posterior: %g", max_lposterior);
		}

		if ((si + 1) % use_every_nth == 0) {
			EmitSample(burnin, sampleIx);

			if (sampleIx + 1 == burnin_period) {
				LOG("End of burnin period");
				LogStatistics();
			}
			
			if (adapt_proposal_samples > 0 && ((sampleIx + 1) % adapt_proposal_samples == 0)) {
				if (sampleIx + 1 != burnin_period) {
					LogStatistics();
				}

				LOG("Updating proposal to empirical variance");
				for (size_t i = 0; i < num_chains; i++) {
					chains[i].AdaptProposal();
				}
			}
		}
	}

	UpdateProgress(1.0);
	LOG("Finished!");

	LogStatistics();

	// Finish calculating the thermodynamic integrand variance
	for (size_t ci = 0; ci < num_chains; ci++) {
		thermodynamic_integrand_variance(ci) /= (sampling_period - 1.0);
	}

	// Calculate evidence
	CalculateMarginalLikelihood();

	if (!output_path.empty()) {
		std::string diag_fn = output_path + "diag_thermodynamic_integration.tsv";
		FILE* file = fopen(diag_fn.c_str(), "w");
		if (file) {
			fputs("temperature\tintegrand\tvariance\n", file);
			for (size_t ci = 0; ci < num_chains; ci++) {
				fprintf(file, "%g\t%g\t%g\n", temperatures(ci), thermodynamic_integrand(ci), thermodynamic_integrand_variance(ci));
			}
			fclose(file);
		} else {
			LOGERROR("Could not open file \"%s\" for writing thermodynamic integration diagnostics", diag_fn.c_str());
		}
	}

#if 0
	DumpCurrentState((OutputPath + "end_state.tsv").c_str());
	DumpVariances((OutputPath + "end_variance.tsv").c_str());
#endif

	AdaptTemperatures();
	
	if (!output_path.empty()) {
		std::string diag_fn = output_path + "suggested_temperatures.tsv";
		FILE* file = fopen(diag_fn.c_str(), "w");
		if (file) {
			fputs("temperature\n", file);
			for (size_t ci = 0; ci < num_chains; ci++) {
				fprintf(file, "%g\n", temperatures(ci));
			}
			fclose(file);
		} else {
			LOGERROR("Could not open file \"%s\" for writing suggested temperatures", diag_fn.c_str());
		}
	}
	
	return RUN_FINISHED;
}

void SamplerPT::DoExchangeMove(bool burnin, size_t sampleIx)
{
	// Randomly pick a pair of neighboring chains
	unsigned long pair = rng->GetUniformInt((unsigned long)(num_chains - 1));
	
	// Attempt to swap these two chains
	SamplerPTChain& chain1 = chains[pair];
	SamplerPTChain& chain2 = chains[pair+1];
	bool exchanged = chain1.ExchangeMove(burnin, chain2);
	
	for (size_t ci = 0; ci < num_chains; ci++) {
		up_labeled_time[ci]++;
		down_labeled_time[ci]++;
	}

	if (exchanged) {
		// Swap labels
		std::swap(labels[pair], labels[pair+1]);
		std::swap(up_labeled_time[pair], up_labeled_time[pair+1]);
		std::swap(down_labeled_time[pair], down_labeled_time[pair+1]);

		if (pair == 0 && labels[0] == 2) {
			up_trip_time.push_back(up_labeled_time[0]);
		} else if(pair+1 == num_chains-1 && labels[pair+1] == 1) {
			down_trip_time.push_back(down_labeled_time[num_chains-1]);
		}
		
		// Relabel top and bottom
		labels[0] = 1;
		labels[num_chains-1] = 2;
		up_labeled_time[0] = 0;
		down_labeled_time[num_chains-1] = 0;
	}
	
	// Update histograms
	for (size_t ci = 0; ci < num_chains; ci++) {
		if (labels[ci] == 2) {
			up_count[ci]++;
		} else if (labels[ci] == 1) {
			down_count[ci]++;
		}
	}
}

bool SamplerPT::DoMutateMove(bool burnin)
{
	bool result = true;
	for (size_t ci = 0; ci < num_chains; ci++) {
		chains[ci].MutateMoveAsync(burnin);
	}
	for (size_t ci = 0; ci < num_chains; ci++) {
		result &= chains[ci].MutateMoveWait();
	}
	return result;
}

void SamplerPT::EmitSample(bool burnin, size_t sampleIx)
{
	if (!burnin) {
		if (posterior_sample_handler) {
			SamplerPTChain& c = chains[num_chains-1];
			posterior_sample_handler(c.GetVarValues().data(), c.GetLogPowerPosterior(), c.GetLogLikelihood(), c.GetLogPrior(), 1.0 / (Real)sampling_period, posterior_sample_handler_user);
		}
		if (output_values_handler && num_output_values > 0) {
			SamplerPTChain& c = chains[num_chains-1];
			output_values_handler(c.GetOutputValues().data(), output_values_handler_user);
		}

		// Update thermodynamic integral
		Real sample_ix_real = static_cast<Real>(sampleIx - burnin_period);
		for (size_t ci = 0; ci < num_chains; ci++) {
			Real prev_mean = thermodynamic_integrand[ci];
			Real x = chains[ci].GetLogLikelihood();
			thermodynamic_integrand[ci] += (x - thermodynamic_integrand[ci]) / (sample_ix_real + 1.0);
			thermodynamic_integrand_variance[ci] += (x - thermodynamic_integrand[ci]) * (x - prev_mean);
		}

		// Store the llh, we need to do further processing with this.
		for (size_t ci = 0; ci < num_chains; ci++) {
			samples_lposterior[ci](sampleIx - burnin_period) = chains[ci].GetLogPowerPosterior();
			//samples_lposterior[ci](sampleIx - burnin_period) = chains[ci].GetVarValues()(0);
		}
	}
}

void SamplerPT::LogStatistics()
{
	// Statistics
	LOG("");
	LOG("Acceptance statistics:");
	LOG("Temperature | Accept (all) | Exchange (all)");
	for (size_t ci = 0; ci < num_chains; ci++) {
		chains[ci].LogStatistics();
	}
	
	LOG("");
	LOG("Proposal info:");
	chains[num_chains-1].LogProposalInfo();

	LOG("");
	LOG("Diffusion statistics:");
	LOG("Temperature | Labels | Upwards count | Downwards count | f");
	for (size_t ci = 0; ci < num_chains; ci++) {
		LOG("%11.7f | %6d | %13.0f | %15.0f | %.7f", temperatures[ci], labels[ci], up_count[ci], down_count[ci], up_count[ci] / (up_count[ci] + down_count[ci]));
	}

	Real avg_upward_rtt = mean(up_trip_time);
	Real avg_downward_rtt = mean(down_trip_time);
	LOG("");
	LOG("Average   upward trip time: %.0f (%u trips)", avg_upward_rtt, up_trip_time.size());
	LOG("Average downward trip time: %.0f (%u trips)", avg_downward_rtt, down_trip_time.size());
}

void SamplerPT::AdaptTemperatures()
{
	// Prevent divisions by zero.
	for (size_t ci = 0; ci < num_chains; ci++) {
		if (up_count[ci] == 0) {
			up_count[ci] = 1;
		}
		if (down_count[ci] == 0) {
			down_count[ci] = 1;
		}
	}
			
	// f(T_i) = nu(T_i) / ( nu(T_i) + nd(T_i) )
	VectorReal f = up_count.cwiseQuotient(up_count + down_count);
	VectorReal temps = temperatures;
			
	// Enforce strictly increasing f
	for (size_t ci = 1; ci < num_chains; ci++) {
		if (f[ci] <= f[ci-1]) {
			f[ci] = f[ci-1] + 1e-10;
		}
	}
	f[0] = 0;
	f[num_chains-1] = 1;

	// Smoothen f (as presented in Hamze et al 2010, Robust parameter selection for parallel tempering).
	const Real w = 0.25;
	VectorReal L = VectorReal::Zero(num_chains);
	for (size_t ci = 0; ci < num_chains; ci++) {
		L(ci) = ci / (Real)(num_chains - 1);
	}
	f = (1 - w) * f + w * L;

	// Estimate df/dT using linear regression
	VectorReal dfdT(num_chains-1);
	int n = 1;
	for (size_t ci = 0; ci < num_chains-1; ci++) {
		int start = std::max(0, (int)ci - n);
		int end = std::min((int)num_chains-1, (int)ci + n);
		int size = end - start + 1;

		VectorReal regx = temps.segment(start, size);
		VectorReal regy = f.segment(start, size);

		Real mx = regx.sum() / regx.size();
		Real my = regy.sum() / regy.size();
		VectorReal mxv = regx.Constant(regx.size(), mx);
		VectorReal myv = regy.Constant(regy.size(), my);
				
		VectorReal dxdx = (regx - mxv).cwiseProduct((regx - mxv));
		VectorReal dxdy = (regx - mxv).cwiseProduct((regy - myv));

		Real b = dxdy.sum() / dxdx.sum();
		dfdT[ci] = b;
	}

	// Calculate eta
	VectorReal eta(num_chains-1);
	for (size_t ci = 0; ci < num_chains-1; ci++) {
		eta[ci] = sqrt(dfdT[ci] / (temps[ci+1] - temps[ci]));
	}

	// Calculate normalization factor C
	Real C = 0.0;
	for (size_t ci = 0; ci < num_chains-1; ci++) {
		C += eta[ci] * (temps[ci+1] - temps[ci]);
	}
	eta /= C;

	// Calculate new temperature T_ci for each chain except chain 0 and N-1.
	VectorReal newtemps = temps;
	for (size_t ci = 1; ci < num_chains - 1; ci++) {
		Real k = ci / (Real)(num_chains-1);

		// Integrate eta from T_1 until T_ci
		Real integrand = 0.0;
		for (size_t i = 0; i < num_chains; i++) {
			if (i == num_chains-1) {
				Real alpha = (k - integrand) / 1.0;
				newtemps[ci] = temps[i] + alpha * (1.0 - temps[i]);
			} else {
				Real next = eta[i] * (temps[i+1] - temps[i]);
				if (integrand + next >= k) {
					Real alpha = (k - integrand) / next;
					newtemps[ci] = temps[i] + alpha * (temps[i+1] - temps[i]);
					break;
				} else {
					integrand += next;
					if (i == num_chains - 1) {
						newtemps[ci] = temps[i] - std::numeric_limits<Real>::epsilon() * (num_chains - ci - 1);
					}
				}
			}
		}
	}

	temperatures = newtemps;
	for (size_t ci = 0; ci < num_chains; ci++) {
		chains[ci].SetTemperature(temperatures[ci]);
	}

	// Reset counts
	up_count.setZero();
	down_count.setZero();
	labels.setZero();
	labels[0] = 1;
	labels[num_chains - 1] = 2;
	up_labeled_time.setZero();
	down_labeled_time.setZero();
	up_trip_time.clear();
	down_trip_time.clear();
}

void SamplerPT::CalculateMarginalLikelihood()
{
	evidence = 0.0;
	Real sampling_error = 0.0;
	Real discretization_error = 0.0;

	// Calculate decorrelation time = the lag at which acf < 2.0 / sqrt(n).
	LOG("Autocorrelation statistics:");
	VectorReal decorrelation_time = VectorReal::Zero(num_chains);
	Real threshold = 2.0 / sqrt((Real)sampling_period);
	for (size_t ci = 0; ci < num_chains; ci++) {
		Real mu = mean(samples_lposterior[ci]);
		Real sigmaSq = var(samples_lposterior[ci]);

		decorrelation_time(ci) = sampling_period / 4.0;
		for (size_t lag = 1; lag < sampling_period / 4; lag++) {
			Real c = acf(samples_lposterior[ci], (int)lag, mu, sigmaSq);
			if (c < threshold) {
				decorrelation_time(ci) = static_cast<Real>(lag);
				break;
			}
		}
		LOG("Decorrelation time chain %4u: %8g", ci, decorrelation_time(ci));
	}

	for (size_t ci = 1; ci < num_chains; ci++) {
		// Trapezoidal integration
		Real deltat = temperatures(ci) - temperatures(ci-1);
		evidence				+= 0.5 * deltat * (thermodynamic_integrand(ci) + thermodynamic_integrand(ci-1));
		discretization_error	+= 0.5 * deltat * (thermodynamic_integrand(ci) - thermodynamic_integrand(ci-1));
		
		// Use the effective sample size instead of the normal sample size for calculating the SEM
		// as suggested in Lartillot & Philippe 2006.
		const Real sem_factor1 = 1.0 / sqrt(sampling_period / decorrelation_time(ci));
		const Real sem_factor2 = 1.0 / sqrt(sampling_period / decorrelation_time(ci-1));
		const Real vi1 = thermodynamic_integrand_variance(ci);
		const Real vi2 = thermodynamic_integrand_variance(ci-1);
		sampling_error			+= 0.25 * deltat * deltat * (vi1 * sem_factor1 + vi2 * sem_factor2);
	}

	LOG("Thermodynamic integration sampling variance: %g", sampling_error);
	LOG("Thermodynamic integration worst-case discretization error: %g", discretization_error);

	// Don't include the discretization error; it
	evidence_sd = sqrt(sampling_error) + discretization_error;
}

void SamplerPT::UpdateProgress(Real p)
{
	if (progress) {
		progress->ReportProgress(p);
	}
}

}
