#include "Utils.h"
#include "Prior.h"
#include "ProgressIndicator.h"
#include "RNG_SFMT.h"
#include "SamplerSMC.h"
#include "SMCSamplingStrategyBlockedMCMC.h"
#include "SMCSamplingStrategyClusteredMCMC.h"
#include "SMCSamplingStrategyKDE.h"
#include "SMCSamplingStrategyMCMC.h"

namespace bcm {

static const size_t NUM_PARTICLES_PER_TASK = 2;

SamplerSMC::SamplerSMC(size_t num_threads)
	: Sampler(num_threads)
	, start_temperature(0.0)
	, end_temperature(1.0)
	, population_size(1000)
	, use_every_nth(1)
	, temperature_schedule(TS_AdaptiveFromESS)
	, schedule_num_populations(500)
	, schedule_ess_ratio(0.9)
	, resampling_ess_threshold(0.5)
	, resampling_strategy(RS_Residual)
	, sampling_strategy_str("blocked_mcmc")
	, mcmc_num_steps(1)
	, mcmc_recalculate(100)
	, mcmc_local_num_nearby(100)
	, mcmc_max_correlation(1.0)
	, final_ess_threshold(0.0)
	, run_until_decorrelation(false)
	, sampling_strategy(NULL)
	, rng(NULL)
	, current_population(NULL)
	, previous_population(NULL)
	, current_iteration(0)
	, population_ix(0)
	, current_temperature(0.0)
	, current_ess(0.0)
	, previous_temperature(0.0)
{
}

SamplerSMC::~SamplerSMC()
{
	for (std::vector<AsyncParams>::iterator ti = async.begin(); ti != async.end(); ++ti) {
		safe_delete(ti->rng);
	}
	safe_delete(sampling_strategy);
	safe_delete(rng);
}

bool SamplerSMC::ParseImpl(const boost::program_options::variables_map& vm)
{
	std::string temperature_schedule_str;

	try {
		population_size				= vm["sampler.population_size"].as<size_t>();
		use_every_nth				= vm["sampler.use_every_nth"].as<size_t>();
		start_temperature			= vm["sampler.start_temperature"].as<Real>();
		end_temperature				= vm["sampler.end_temperature"].as<Real>();
		temperature_schedule_str	= vm["sampler.temperature_schedule"].as<std::string>();
		schedule_ess_ratio			= vm["sampler.temperature_schedule_adaptive_ratio"].as<Real>();
		schedule_num_populations	= vm["sampler.temperature_schedule_fixed_number"].as<size_t>();
		resampling_ess_threshold	= vm["sampler.resampling_threshold"].as<Real>();
		sampling_strategy_str		= vm["sampler.smc_strategy"].as<std::string>();
		mcmc_num_steps				= vm["sampler.smc_mcmc_steps"].as<size_t>();
		mcmc_recalculate			= vm["sampler.smc_mcmc_recalculate"].as<size_t>();
		mcmc_local_num_nearby		= vm["sampler.smc_mcmc_local_num_nearby"].as<size_t>();
		mcmc_max_correlation		= vm["sampler.smc_mcmc_max_correlation"].as<Real>();
		final_ess_threshold			= vm["sampler.smc_final_ess_threshold"].as<Real>();
		run_until_decorrelation		= vm["sampler.smc_run_until_decorrelation"].as<bool>();
	} catch (boost::program_options::error& e) {
		LOGERROR("Error parsing sampler: %s", e.what());
		return false;
	}

	std::transform(temperature_schedule_str.begin(), temperature_schedule_str.end(), temperature_schedule_str.begin(), ::tolower);

	// Make sure all of the options have reasonable values.
	if (population_size == 0) {
		LOGERROR("population_size must be greater than 0");
		return false;
	}

	if (start_temperature < 0 || start_temperature > 1) {
		LOGERROR("start_temperature must be between 0 and 1 inclusive");
		return false;
	}

	if (end_temperature < 0 || end_temperature > 1 || end_temperature < start_temperature) {
		LOGERROR("end_temperature must be between 0 and 1 inclusive, and must be >= start_temperature");
		return false;
	}

	if (temperature_schedule_str == "fixed" || temperature_schedule_str == "fixed_linear") {
		temperature_schedule = TS_FixedLinear;
	} else if (temperature_schedule_str == "fixed_loglinear") {
		temperature_schedule = TS_FixedLogLinear;
	} else if (temperature_schedule_str == "adaptive") {
		temperature_schedule = TS_AdaptiveFromESS;
	} else {
		LOGERROR("Unknown temperature_schedule \"%s\"", temperature_schedule_str.c_str());
		return false;
	}

	if ((temperature_schedule == TS_FixedLinear || temperature_schedule == TS_FixedLogLinear) && schedule_num_populations == 0) {
		LOGERROR("schedule_num_populations must be greater than 0");
		return false;
	}
	if (temperature_schedule == TS_AdaptiveFromESS && (schedule_ess_ratio < 0 || schedule_ess_ratio > 1)) {
		LOGERROR("schedule_ess_ratio must be between 0 and 1 inclusive");
		return false;
	}

	if (resampling_ess_threshold < 0 || resampling_ess_threshold > 1) {
		LOGERROR("resampling_ess_threshold must be between 0 and 1 inclusive");
		return false;
	}
	
	if (sampling_strategy_str == "mcmc_global" || sampling_strategy_str == "mcmc_local") {
		if (mcmc_num_steps == 0) {
			LOGERROR("mcmc_num_steps must be at least 1");
			return false;
		}
	}
	if (sampling_strategy_str == "mcmc_local") {
		if (mcmc_local_num_nearby == 0) {
			LOGERROR("mcmc_local_num_nearby must be at least 1");
			return false;
		}
	}

	return true;
}

bool SamplerSMC::InitializeImpl()
{
	rng = new RNG_SFMT(rngseed);

	// Initialize populations
	populations[0].resize(population_size);
	populations[1].resize(population_size);
	
	correlation_work_vector[0].resize(population_size);
	correlation_work_vector[1].resize(population_size);
	correlation_work_vector[2].resize(population_size);

	for (size_t i = 0; i < population_size; i++) {
		populations[0][i].var_values.resize(num_variables);
		populations[1][i].var_values.resize(num_variables);
	}
	current_iteration = 0;
	population_ix = 0;
	current_population = &populations[0];
	previous_population = NULL;

	current_temperature = start_temperature;

	if (output_values_handler && num_output_values > 0) {
		for (size_t i = 0; i < population_size; i++) {
			populations[0][i].output_values.resize(num_output_values);
			populations[1][i].output_values.resize(num_output_values);
		}
	}
	
	// Initialize sampling strategy
	if (sampling_strategy_str == "mcmc_global") {
		SMCSamplingStrategyMCMC* strat = new SMCSamplingStrategyMCMC(this, task_manager);
		strat->SetGlobal();
		strat->SetRecalculateSteps(mcmc_recalculate);
		strat->SetStepsPerIteration(mcmc_num_steps);
		sampling_strategy = strat;
	} else if (sampling_strategy_str == "mcmc_global_covariance") {
		SMCSamplingStrategyMCMC* strat = new SMCSamplingStrategyMCMC(this, task_manager);
		strat->SetGlobalCovariance();
		strat->SetRecalculateSteps(mcmc_recalculate);
		strat->SetStepsPerIteration(mcmc_num_steps);
		sampling_strategy = strat;
	} else if (sampling_strategy_str == "mcmc_local") {
		size_t num_nearby_particles = mcmc_local_num_nearby * num_variables;
		num_nearby_particles = std::min(num_nearby_particles, population_size);
		num_nearby_particles = std::max(num_nearby_particles, (size_t)10);

		SMCSamplingStrategyMCMC* strat = new SMCSamplingStrategyMCMC(this, task_manager);
		strat->SetLocal(num_nearby_particles);
		strat->SetRecalculateSteps(mcmc_recalculate);
		strat->SetStepsPerIteration(mcmc_num_steps);
		sampling_strategy = strat;
	} else if (sampling_strategy_str == "blocked_mcmc") {
		SMCSamplingStrategyBlockedMCMC* strat = new SMCSamplingStrategyBlockedMCMC(this, task_manager);
		strat->SetRecalculateSteps(mcmc_recalculate);
		strat->SetStepsPerIteration(mcmc_num_steps);
		sampling_strategy = strat;
	} else if (sampling_strategy_str == "kde_local") {
		sampling_strategy = new SMCSamplingStrategyKDE(this, task_manager);
	} else if (sampling_strategy_str == "clustered_mcmc") {
		SMCSamplingStrategyClusteredMCMC* strat = new SMCSamplingStrategyClusteredMCMC(this, task_manager);
		strat->SetRecalculateSteps(mcmc_recalculate);
		strat->SetStepsPerIteration(mcmc_num_steps);
		sampling_strategy = strat;
	} else {
		LOGERROR("Unknown sampling strategy \"%s\"", sampling_strategy_str.c_str());
		return false;
	}

	// Async parameters
	async.resize(num_threads);
	for (size_t i = 0; i < num_threads; i++) {
		async[i].rng = new RNG_SFMT(rng->GetInt());
		async[i].failure = false;
	}

	return true;
}

void SamplerSMC::UninitializeImpl()
{
	for (size_t i = 0; i < num_threads; i++) {
		safe_delete(async[i].rng);
	}
	async.clear();

	safe_delete(sampling_strategy);
	safe_delete(rng);

	populations[0].clear();
	populations[1].clear();
	current_population = NULL;
	previous_population = NULL;
	current_iteration = 0;
	population_ix = 0;
}

Sampler::ERunResult SamplerSMC::RunImpl()
{
	bool result;
	bool force_strategy_update = true;
	const Real significant_correlation = std::max(2.0 / sqrt((Real)population_size), 0.1);

	if (multistart_continue) {
		// Update progress indicator
		UpdateProgress(0.0, current_temperature, current_ess, 0.0);
		current_population = &populations[population_ix % 2];
		previous_population = &populations[(population_ix-1) % 2];
	} else {
		// Initial sampling from prior.
		current_temperature = start_temperature;
		result = SampleInitialPopulation();
		if (!result) {
			return RUN_ERROR;
		}
		population_at_resample = *current_population;
		current_ess = CalculateESS();
		LOG("Initial population ESS: %g", current_ess);

		// Contribution to evidence from sampling from the prior
		// Most likely this is very close to zero, unless there are areas of the prior which have a likelihood density of 0.
		evidence = -std::numeric_limits<Real>::infinity();
		evidence_sd = -std::numeric_limits<Real>::infinity();
		CalculateEvidenceRatio(current_population, evidence, evidence_sd);
		LOG("Initial evidence contribution: %g +/- %g", evidence, sqrt(evidence_sd));

		// Update progress indicator
		UpdateProgress(0.0, current_temperature, current_ess, 0.0);
	}

	// Run the sequential sampler according to the chosen temperature schedule.
	bool finished = false;
	Timer timer;
	timer.Start();
	while (1) {
		// Are we finished?
		if (current_temperature == end_temperature) {
			finished = true;
			break;
		}

		// Have we run out of time?
		if (runtime_seconds > 0) {
			Real elapsed = timer.GetElapsedSeconds();
			if (elapsed > runtime_seconds) {
				SaveCurrentState();
				break;
			}
		}

		current_iteration++;

		// Do we need to resample?
		if (current_ess <= resampling_ess_threshold * population_size) {
			LOG("ESS dropped below threshold, resampling...");

			Real r = CalculateCorrelation(current_population, &population_at_resample);
			LOG("Correlation since previous resample step: r = %g", r);
			if (r > significant_correlation) {
				LOG("WARNING - populations still correlated at resample!");
			}
			
			// Update evidence calculation
			Real ratio = 0.0, ratio_sd = 0.0;
			CalculateEvidenceRatio(current_population, ratio, ratio_sd);
			evidence += ratio;
			evidence_sd += ratio_sd;
			LOG("Evidence contribution since resample: %g +/- %g", ratio, sqrt(ratio_sd));

			// Resample
			previous_population = current_population;
			population_ix++;
			current_population = &populations[population_ix % 2];

			Resample();
			force_strategy_update = true;
			population_at_resample = *current_population;

			current_ess = static_cast<Real>(population_size);
		}
		
		// Swap populations
		previous_population = current_population;
		population_ix++;
		current_population = &populations[population_ix % 2];
		
		// Calculate next temperature
		result = CalculateNextTemperature(current_iteration);
		if (!result) {
			return RUN_ERROR;
		}

		// Give the sampling strategy a chance to update its parameters
		result = sampling_strategy->UpdateStrategyIteration(force_strategy_update, rng);
		if (!result) {
			return RUN_ERROR;
		}
		force_strategy_update = false;

		// Sample new population
		result = SampleNewPopulation();
		if (!result) {
			return RUN_ERROR;
		}

		// Normalize weights
		result = NormalizeWeights();
		if (!result) {
			return RUN_ERROR;
		}

		// Calculate some statistics (ESS, correlation and thermodynamic integrand)
		current_ess = CalculateESS();
		Real r = CalculateCorrelation(current_population, previous_population);
		Real ti_integrad = 0, ti_variance = 0;
		CalculateMeanLogLikelihood(current_ess, ti_integrad, ti_variance);

		// Update progress indicator
		if (temperature_schedule == TS_FixedLinear || temperature_schedule == TS_FixedLogLinear) {
			UpdateProgress(current_iteration / (Real)(schedule_num_populations - 1), current_temperature, current_ess, r);
		} else if (temperature_schedule == TS_AdaptiveFromESS) {
			UpdateProgress((current_temperature - start_temperature) / (end_temperature - start_temperature), current_temperature, current_ess, r);
		}
		LOG("t=%g - ESS=%.1f - steps=%u - r=%g - TI=%g +/- %g", current_temperature, current_ess, steps_between_population, r, ti_integrad, ti_variance);
	}

	if (finished) {
		// Final contribution of the remaining samples to the evidence.
		Real ratio = 0.0, ratio_sd = 0.0;
		CalculateEvidenceRatio(current_population, ratio, ratio_sd);
		evidence += ratio;
		evidence_sd += ratio_sd;
		LOG("Final evidence contribution: %g +/- %g", ratio, sqrt(ratio_sd));

		evidence_sd = sqrt(evidence_sd);

		if (current_ess < population_size * final_ess_threshold) {
			LOG("ESS below final threshold, resampling...");

			// Resample
			previous_population = current_population;
			population_ix++;
			current_population = &populations[population_ix % 2];

			Resample();
			force_strategy_update = true;
			population_at_resample = *current_population;
			
			current_ess = static_cast<Real>(population_size);
		}

		if (run_until_decorrelation) {
			LOG("Sampling until final population is decorrelated from previous");
		
			// Swap populations
			previous_population = current_population;
			population_ix++;
			current_population = &populations[population_ix % 2];

			previous_temperature = end_temperature;
			current_temperature = end_temperature;

			result = sampling_strategy->UpdateStrategyIteration(force_strategy_update, rng);
			if (!result) {
				return RUN_ERROR;
			}
			
			// Sample new population until correlation is sufficiently low
			mcmc_max_correlation = significant_correlation;
			result = SampleNewPopulation();
			if (!result) {
				return RUN_ERROR;
			}
		}

		// Output the posterior samples
		for (size_t i = 0; i < population_size; i++) {
			if ((i % use_every_nth) == 0) {
				Particle& particle = (*current_population)[i];
				if (posterior_sample_handler) {
					posterior_sample_handler(particle.var_values.data(), particle.lpowerposterior, particle.llh, particle.lprior, particle.normweight, posterior_sample_handler_user);
				}
				if (output_values_handler && num_output_values > 0) {
					output_values_handler(particle.output_values.data(), output_values_handler_user);
				}
			}
		}

		return RUN_FINISHED;
	} else {
		return RUN_TIMED_END;
	}
}

bool SamplerSMC::SampleInitialPopulation()
{
	// Initial sampling from prior
	LOG("Sampling initial population from prior...");

	for (size_t i = 0; i < num_threads; i++) {
		async[i].logtot = -std::numeric_limits<Real>::infinity();
	}
	
	size_t numtasks = (population_size + NUM_PARTICLES_PER_TASK - 1) / NUM_PARTICLES_PER_TASK;
	async_tasks.resize(numtasks);
	for (size_t i = 0; i < numtasks; i++) {
		TTask task = boost::bind(&SamplerSMC::AsyncDoInitialSample, this, _1, _2);
		async_tasks[i] = task_manager->AddTask(task, reinterpret_cast<void*>(i));
	}
	for (size_t i = 0; i < numtasks; i++) {
		task_manager->WaitTask(async_tasks[i]);
	}

	Real logtot = -std::numeric_limits<Real>::infinity();
	for (size_t i = 0; i < num_threads; i++) {
		if (async[i].failure) {
			return false;
		}
		logtot = logsum(logtot, async[i].logtot);
	}
	
	if (logtot == -std::numeric_limits<Real>::infinity()) {
		LOGERROR("Zero particles left after initial sampling! Consider using a lower starting temperature.");
		return false;
	}
	if (logtot != logtot) {
		LOGERROR("Internal error - NAN in weight calculation");
		return false;
	}
	
	for (size_t j = 0; j < population_size; j++) {
		Particle& particle = (*current_population)[j];
		particle.lnormweight = particle.lweight - logtot;
		particle.normweight = exp(particle.lweight - logtot);
		particle.accumulated_lweight = particle.lnormweight;
	}

	return true;
}

bool SamplerSMC::SampleNewPopulation()
{
	for (size_t j = 0; j < population_size; j++) {
		Particle& new_particle = (*current_population)[j];
		const Particle& previous_particle = (*previous_population)[j];

		new_particle.var_values = previous_particle.var_values;
		new_particle.output_values = previous_particle.output_values;
		new_particle.lprior = previous_particle.lprior;
		new_particle.llh = previous_particle.llh;
		new_particle.lpowerposterior = previous_particle.lprior + current_temperature * previous_particle.llh;
		
		if (previous_particle.normweight <= 0) {
			new_particle.lweight = -std::numeric_limits<Real>::infinity();
		}
	}

	steps_between_population = 0;

	// Update particles until correlation is below threshold
	while (1) {
		steps_between_population++;

		size_t numtasks = (population_size + NUM_PARTICLES_PER_TASK - 1) / NUM_PARTICLES_PER_TASK;
		async_tasks.resize(numtasks);
		for (size_t i = 0; i < numtasks; i++) {
			TTask task = boost::bind(&SamplerSMC::AsyncDoSample, this, _1, _2);
			async_tasks[i] = task_manager->AddTask(task, reinterpret_cast<void*>(i));
		}
		for (size_t i = 0; i < numtasks; i++) {
			task_manager->WaitTask(async_tasks[i]);
		}

		for (size_t i = 0; i < num_threads; i++) {
			if (async[i].failure) {
				return false;
			}
		}
		
		if (mcmc_max_correlation < 1) {
			Real r = CalculateCorrelation(current_population, previous_population);

			if (r < mcmc_max_correlation) {
				break;
			} else {
				UpdateProgress((current_temperature - start_temperature) / (end_temperature - start_temperature), current_temperature, current_ess, r);

				// Run another sampling round, until correlation is below max_correlation.
				// First give the sampling strategy another chance at minor updates
				if (!sampling_strategy->UpdateStrategyCurrent(rng)) {
					return false;
				}
			}
		} else {
			break;
		}
	}

	// Update weights
	// TODO - for KDE strategy this is computationally costly, parallelize.
	for (size_t j = 0; j < population_size; j++) {
		sampling_strategy->CalculateParticleWeight((*current_population)[j], j, 0);
	}

	return true;
}

void SamplerSMC::Resample()
{
	Real new_normweight = 1.0 / (Real)population_size;
	Real lnew_normweight = log(new_normweight);

	if (resampling_strategy == RS_Multinomial) {
		for (size_t i = 0; i < population_size; i++) {
			// Sample a particle from the previous population according to their weights
			Real alpha = rng->GetReal();
			Real t = 0;
			size_t pix = population_size - 1;
			for (size_t j = 0; j < population_size-1; j++) {
				Particle& previous_particle = (*previous_population)[j];
				t += previous_particle.normweight;
				if (alpha < t) {
					pix = j;
					break;
				}
			}

			// Copy that particle
			(*current_population)[i] = (*previous_population)[pix];
		}
	} else if (resampling_strategy == RS_Residual) {
		// First copy over previous samples with relative weight > 1
		size_t ni = 0;
		for (size_t i = 0; i < population_size; i++) {
			Particle& previous_particle = (*previous_population)[i];
			Real relweight = (Real)population_size * previous_particle.normweight;
			size_t copies = (size_t)floor((Real)population_size * previous_particle.normweight);
			
			for (size_t ci = 0; ci < copies; ci++) {
				(*current_population)[ni] = previous_particle;
				ni++;
			}
		}

		// For the remaining particles, do multinomial sampling on the residuals
		Real total_residual_weight = (Real)(population_size - ni);
		for (; ni < population_size; ni++) {
			Real alpha = rng->GetReal() * total_residual_weight;
			Real t = 0;
			size_t pix = population_size - 1;
			for (size_t j = 0; j < population_size-1; j++) {
				Particle& previous_particle = (*previous_population)[j];
				Real relweight = (Real)population_size * previous_particle.normweight;
				Real residual_weight = relweight - floor(relweight);
				
				t += residual_weight;
				if (alpha < t) {
					pix = j;
					break;
				}
			}
			
			(*current_population)[ni] = (*previous_population)[pix];
		}
	}

	// Set the weights
	for (size_t i = 0; i < population_size; i++) {
		Particle& new_particle = (*current_population)[i];
		new_particle.lweight = lnew_normweight;
		new_particle.normweight = new_normweight;
		new_particle.lnormweight = lnew_normweight;
		new_particle.accumulated_lweight = lnew_normweight;//(*previous_population)[pix].lnormweight;
	}
}

void SamplerSMC::AsyncDoInitialSample(void* user, size_t threadix)
{
	size_t ti = reinterpret_cast<size_t>(user);

	size_t start_particle = ti * NUM_PARTICLES_PER_TASK;
	size_t end_particle = start_particle + NUM_PARTICLES_PER_TASK;
	end_particle = std::min(end_particle, population_size);

	for (size_t j = start_particle; j < end_particle; j++) {
		Particle& particle = (*current_population)[j];

		if (!prior->Sample(particle.var_values.data(), async[threadix].rng)) {
			LOGERROR("Sampling from prior failed");
			async[threadix].failure = true;
			return;
		}

		if (!EvaluatePriorLikelihood(threadix, particle.var_values.data(), particle.lprior, particle.llh)) {
			async[threadix].failure = true;
			return;
		}

		if (output_values_handler && num_output_values > 0) {
			if (!GetOutputValues(threadix, particle.output_values.data())) {
				async[threadix].failure = true;
				return;
			}
		}

		if (particle.llh == -std::numeric_limits<Real>::infinity()) {
			particle.lpowerposterior = -std::numeric_limits<Real>::infinity();
			particle.lweight = -std::numeric_limits<Real>::infinity();
		} else {
			particle.lpowerposterior = particle.lprior + current_temperature * particle.llh;
			particle.lweight = particle.lpowerposterior - particle.lprior;
		}

		async[threadix].logtot = logsum(async[threadix].logtot, particle.lweight);
	}
}

void SamplerSMC::AsyncDoSample(void* user, size_t threadix)
{
	size_t ti = reinterpret_cast<size_t>(user);

	size_t start_particle = ti * NUM_PARTICLES_PER_TASK;
	size_t end_particle = start_particle + NUM_PARTICLES_PER_TASK;
	end_particle = std::min(end_particle, population_size);

	for (size_t i = start_particle; i < end_particle; i++) {
		Particle& particle = (*current_population)[i];

		if (particle.lweight != -std::numeric_limits<Real>::infinity()) {
			// Let the sampling strategy update the particle
			bool result = sampling_strategy->UpdateParticle(particle, i, async[threadix].rng, threadix);
			if (!result) {
				async[threadix].failure = true;
				return;
			}
		} else {
		}
	}
}

void SamplerSMC::UpdateProgress(Real p, Real t, Real ess, Real r)
{
	if (progress) {
		char buf[16];

		snprintf(buf, sizeof(buf), "%6g", t);
		progress->UpdateInfo("t", buf);
		
		snprintf(buf, sizeof(buf), "%.1f", ess);
		progress->UpdateInfo("ess", buf);
		
		snprintf(buf, sizeof(buf), "%.6f", r);
		progress->UpdateInfo("corr", buf);

		progress->ReportProgress(p);
	}
}

}
