#include "Utils.h"
#include "MathFunctions.h"

namespace bcm {

Real sum(const std::vector<Real>& x)
{
	Real sum = 0.0;
	for (size_t i = 0; i < x.size(); ++i) {
		sum += x[i];
	}
	return sum;
}

Real mean(const std::vector<Real>& x)
{
	Real mu = 0.0;
	for (size_t i = 0; i < x.size(); ++i) {
		mu += (x[i] - mu) / (i + 1);
	}
	return mu;
}

Real var(const std::vector<Real>& x)
{
	const Real mu = mean(x);
	return var(x, mu);
}

Real var(const std::vector<Real>& x, const Real mu)
{
	Real sigmaSq = 0.0;
	for (size_t i = 0; i < x.size(); ++i) {
		const Real d = x[i] - mu;
		sigmaSq += (d*d - sigmaSq) / (i+1);
	}
	const Real n = static_cast<Real>(x.size());
	return sigmaSq * (n / (n - 1.0));
}

Real acf(const std::vector<Real>& x, const size_t lag)
{
	if (lag == 0) {
		return 1.0;
	}
	if (x.size() <= lag) {
		return std::numeric_limits<Real>::quiet_NaN();
	}

	const Real mu = mean(x);
	const Real sigmaSq = var(x, mu);

	Real r = 0.0;
	for (size_t i = 0; i < x.size() - lag; i++) {
		const Real x1 = x[i] - mu;
		const Real x2 = x[i+lag] - mu;
		r += (x1*x2 - r) / (i+1);
	}

	return r / sigmaSq;
}

Real sum(const VectorReal& x)
{
	Real sum = 0.0;
	for (int i = 0; i < x.size(); ++i) {
		sum += x[i];
	}
	return sum;
}

Real mean(const VectorReal& x)
{
	Real mu = 0.0;
	for (int i = 0; i < x.size(); ++i) {
		mu += (x[i] - mu) / (i + 1);
	}
	return mu;
}

Real var(const VectorReal& x)
{
	const Real mu = mean(x);
	return var(x, mu);
}

Real var(const VectorReal& x, const Real mu)
{
	Real sigmaSq = 0.0;
	for (int i = 0; i < x.size(); ++i) {
		const Real d = x[i] - mu;
		sigmaSq += (d*d - sigmaSq) / (i+1);
	}
	const Real n = static_cast<Real>(x.size());
	return sigmaSq * (n / (n - 1.0));
}

Real acf(const VectorReal& x, const int lag)
{
	if (lag == 0) {
		return 1.0;
	}
	if (x.size() <= lag) {
		return std::numeric_limits<Real>::quiet_NaN();
	}

	const Real mu = mean(x);
	const Real sigmaSq = var(x, mu);

	Real r = 0.0;
	for (int i = 0; i < x.size() - lag; i++) {
		const Real x1 = x[i] - mu;
		const Real x2 = x[i+lag] - mu;
		r += (x1*x2 - r) / (i+1);
	}

	return r / sigmaSq;
}

Real acf(const VectorReal& x, const int lag, const Real mu, const Real sigmaSq)
{
	if (lag == 0) {
		return 1.0;
	}
	if (x.size() <= lag) {
		return std::numeric_limits<Real>::quiet_NaN();
	}

	Real r = 0.0;
	for (int i = 0; i < x.size() - lag; i++) {
		const Real x1 = x[i] - mu;
		const Real x2 = x[i+lag] - mu;
		r += (x1*x2 - r) / (i+1);
	}

	return r / sigmaSq;
}

VectorReal mean(const std::vector<VectorReal>& x)
{
	if (x.empty()) {
		return VectorReal();
	}

	VectorReal mu = VectorReal::Zero(x[0].size());
	for (size_t i = 0; i < x.size(); ++i) {
		const Real invip1 = 1.0 / (i+1);
		for (int j = 0; j < x[i].size(); j++) {
			mu[j] += (x[i][j] - mu[j]) * invip1;
		}
	}
	return mu;
}

VectorReal var(const std::vector<VectorReal>& x)
{
	const VectorReal mu = mean(x);
	return var(x, mu);
}

VectorReal var(const std::vector<VectorReal>& x, const VectorReal& mu)
{
	VectorReal sigmaSq = VectorReal::Zero(x[0].size());
	for (size_t i = 0; i < x.size(); ++i) {
		const Real invip1 = 1.0 / (i+1);
		const VectorReal d = x[i] - mu;
		for (int j = 0; j < x[i].size(); j++) {
			sigmaSq[j] += (d[j]*d[j] - sigmaSq[j]) * invip1;
		}
	}
	const Real n = static_cast<Real>(x.size());
	return sigmaSq * (n / (n - 1.0));
}

VectorReal acf(const std::vector<VectorReal>& x, const size_t lag)
{
	if (x.size() <= lag) {
		return VectorReal();
	}
	if (lag == 0) {
		return VectorReal::Ones(x[0].size());
	}

	const VectorReal mu = mean(x);
	const VectorReal sigmaSq = var(x, mu);

	VectorReal r = VectorReal::Zero(x[0].size());
	for (size_t i = 0; i < x.size() - lag; i++) {
		const Real invip1 = 1.0 / (i+1);
		for (int j = 0; j < x[i].size(); j++) {
			const Real x1 = x[i][j] - mu[j];
			const Real x2 = x[i+lag][j] - mu[j];
			r[j] += (x1*x2 - r[j]) * invip1;
		}
	}

	return r.cwiseQuotient(sigmaSq);
}

}
