#pragma once

#include <boost/math/special_functions/log1p.hpp>

namespace bcm {

Real sum(const std::vector<Real>& x);
Real mean(const std::vector<Real>& x);
Real var(const std::vector<Real>& x);
Real var(const std::vector<Real>& x, const Real mu);
Real acf(const std::vector<Real>& x, const size_t lag);

Real sum(const VectorReal& x);
Real mean(const VectorReal& x);
Real var(const VectorReal& x);
Real var(const VectorReal& x, const Real mu);
Real acf(const VectorReal& x, const int lag);
Real acf(const VectorReal& x, const int lag, const Real mu, const Real sigmaSq);

VectorReal mean(const std::vector<VectorReal>& x);
VectorReal var(const std::vector<VectorReal>& x);
VectorReal var(const std::vector<VectorReal>& x, const VectorReal& mu);
VectorReal acf(const std::vector<VectorReal>& x, const size_t lag);

inline float log2(float x) { return 1.4426950408889634073599246810019f * logf(x); }
inline double log2(double x) { return 1.4426950408889634073599246810019 * log(x); }
inline float log10(float x) { return 0.4342944819032518276511289189166f * logf(x); }
inline double log10(double x) { return 0.4342944819032518276511289189166 * log(x); }
inline float fastpow10(float x) { return exp(x * 2.3025850929940459f); }
inline double fastpow10(double x) { return exp(x * 2.3025850929940459); }

inline Real logistic_activation(Real x, Real steepness, Real inflection)
{
	return 1.0 / (1.0 + exp(-steepness * (x - inflection)));
}

inline Real logistic_activation_scaled(Real x, Real steepness, Real inflection)
{
#if 0
	Real y0 = 1.0 / (1.0 + exp(steepness * inflection));
	Real y1 = 1.0 / (1.0 + exp(-steepness * (1.0 - inflection)));
	Real yx = 1.0 / (1.0 + exp(-steepness * (x - inflection)));
	return (yx - y0) / (y1 - y0);
#else
	// Rearranged to save a few divisions
	Real exps = exp(steepness);
	Real expsx = exp(steepness * x);
	Real expinfs = exp(inflection * steepness);
	return (
			((expinfs+exps) * (expsx-1)) / 
			((exps-1) * (expinfs+expsx))
		   );
#endif
}

inline Real logistic(Real x) { return 1.0 / (1.0 + exp(-x)); }
inline Real logit(Real x) { return log(x / (1.0 - x)); }

inline Real logistic_activation_normal(Real x, Real steepness, Real inflection, Real max)
{
	Real S = pow(100, steepness);
	Real y0 = 1.0 / (1.0 + exp(S * inflection));
	Real y1 = 1.0 / (1.0 + exp(-S * (1.0 - inflection)));
	Real yx = 1.0 / (1.0 + exp(-S * (x - inflection)));
	return max * (yx - y0) / (y1 - y0);
}

inline Real hill_function(Real x, Real n, Real k, Real max)
{
	return max / (pow(k / x, n) + 1);
}

inline Real hill_function_scaled(Real x, Real log20n, Real k, Real max)
{
	Real n = pow(20, log20n);
	return max / (pow(k / x, n) + 1);
}

inline float rsqrt(float x)
{
	// Based on http://www.hlnum.org/english/doc/frsqrt/frsqrt.html
	__m128 f = _mm_set_ss(x);
	f = _mm_rsqrt_ss(f);
	float r = _mm_cvtss_f32(f);
	r *= ((3.0f - r * r * x) * 0.5f);
	return r;
}

inline double rsqrt(double x)
{
	// Based on http://www.hlnum.org/english/doc/frsqrt/frsqrt.html
	__m128 f = _mm_set_ss((float)x);
	f = _mm_rsqrt_ss(f);
	double r = (double)_mm_cvtss_f32(f);
	r *= ((3.0 - r * r * x) * 0.5);
	r *= ((3.0 - r * r * x) * 0.5);
	return r;
}

inline float logsumf(float loga, float logb)
{
	if (logb > loga) {
		std::swap(loga, logb);
	}
	if (loga == -std::numeric_limits<float>::infinity()) {
		return loga;
	}

	float diff = logb - loga;
	if (diff < -60) {
		return loga;
	}
	float expdiff = exp(diff);
	return loga + boost::math::log1p(expdiff);
}

inline double logsum(double loga, double logb)
{
	if (logb > loga) {
		std::swap(loga, logb);
	}
	if (loga == -std::numeric_limits<double>::infinity()) {
		return loga;
	}

	double diff = logb - loga;
	if (diff < -500) {
		return loga;
	}
	double expdiff = exp(diff);
	return loga + boost::math::log1p(expdiff);
}

inline double logaminusb(double loga, double logb)
{
	if (loga == -std::numeric_limits<double>::infinity()) {
		return loga;
	}
	if (logb == -std::numeric_limits<double>::infinity()) {
		return loga;
	}
	if (logb > loga) {
		return std::numeric_limits<double>::quiet_NaN();
	}

	double diff = logb - loga;
	if (diff > 500) {
		return 500;
	} else {
		double expdiff = exp(diff);
		return loga + boost::math::log1p(-expdiff);
	}
}

inline double logabsaminusb(double loga, double logb)
{
	if (logb > loga) {
		std::swap(loga, logb);
	}
	if (loga == -std::numeric_limits<double>::infinity()) {
		return logb;
	}
	if (logb == -std::numeric_limits<double>::infinity()) {
		return loga;
	}

	double diff = logb - loga;
	if (diff > 500) {
		return logb;
	} else {
		double expdiff = exp(diff);
		return loga + boost::math::log1p(-expdiff);
	}
}

inline int factorial(int a)
{
	int result = 1;
	for (int i = 2; i <= a; i++) {
		result *= i;
	}
	return result;
}

inline Real logfactorial_approx(Real a)
{
	//return a * log(a) - a;
	return a * log(a) - a + log(a * (1 + 4 * a * (1 + 2 * a))) / 6.0 + 0.57236494292470008707171367567653;
}

}
