#include "Utils.h"
#include "TaskManager.h"
#include <boost/interprocess/detail/atomic.hpp>

namespace bcm {

void worker(TaskManager* manager, size_t threadIndex)
{
	manager->WorkerFunction(threadIndex);
}

TaskManager::TaskManager(size_t numthreads)
	: NextHandle(0)
	, Active(true)
{
	Threads.resize(numthreads);
	for (size_t i = 0; i < numthreads; i++) {
		Threads[i] = new boost::thread(worker, this, i);
	}
}

TaskManager::~TaskManager()
{
	{
		boost::lock_guard<boost::mutex> lock(StartMutex);
		Active = false;
		StartCondition.notify_all();
	}
	
	for (size_t i = 0; i < Threads.size(); i++) {
		Threads[i]->join();
		delete Threads[i];
	}
}

uint64 TaskManager::AddTask(TTask task, void* userdata)
{
	boost::lock_guard<boost::mutex> lock(StartMutex);
	Task t;
	t.Delegate = task;
	t.UserData = userdata;
	t.Handle = NextHandle;
	NextHandle++;
	WaitingTasks.push_back(t);
	StartCondition.notify_one();
	return t.Handle;
}

void TaskManager::WaitTask(uint64 task)
{
	boost::unique_lock<boost::mutex> lock(FinishMutex);
    while(1) {
		bool finished = false;
		for (std::list<Task>::iterator ti = FinishedTasks.begin(); ti != FinishedTasks.end(); ++ti) {
			if (ti->Handle == task) {
				finished = true;
				FinishedTasks.erase(ti);
				break;
			}
		}

		if (finished) {
			break;
		} else {
			FinishCondition.wait(lock);
		}
    }
}

void TaskManager::WorkerFunction(size_t threadIndex)
{
	while (1) {
		Task task;

		// Wait for a task.
		{
			boost::unique_lock<boost::mutex> lock(StartMutex);
			std::list<Task>::iterator ti;
			while (1) {
				if (!Active) {
					break;
				}
				ti = WaitingTasks.begin();
				if (ti != WaitingTasks.end()) {
					task = *ti;
					WaitingTasks.erase(ti);
					break;
				} else {
					StartCondition.wait(lock);
				}
			}
		}
			
		if (!Active) {
			break;
		}

		// Execute task.
		task.Delegate(task.UserData, threadIndex);

		// Append task to finished lists.
		{
			boost::lock_guard<boost::mutex> lock(FinishMutex);
			FinishedTasks.push_back(task);
		}
		FinishCondition.notify_one();
	}
}

}
